# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Limits on tensor size and decimal representation."""

# Maximum allowable number of elements in a tensor.
MAX_TENSOR_ELEMENTS = 1000

# Maximum allowable number of dimensions for a tensor.
MAX_NUM_DIMENSIONS = 4

# Maximum length of one dimension of a tensor.
MAX_DIMENSION_LENGTH = 100

# The number of floating-point decimal places to consider.
NUM_DECIMALS = 5



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for filtered_values_cache.py."""

import collections

from absl.testing import absltest
from tf_coder.value_search import filtered_values_cache
from tf_coder.value_search import value


def _value(wrapped_value):
  """A simple utility to create Value objects."""
  return value.ConstantValue(wrapped_value)


def _self_mapping(items):
  """Turns a list into an identity mapping dict."""
  return collections.OrderedDict(zip(items, items))


class FilteredValuesCacheTest(absltest.TestCase):

  def test_filter_values(self):
    filter_cache = filtered_values_cache.FilteredValuesCache()
    values_iterable = _self_mapping([_value(2), _value(3), _value(4)])
    even_filter = lambda arg_value: arg_value.value % 2 == 0

    self.assertEqual(
        filter_cache.filter_values(even_filter, 1, values_iterable),
        [_value(2), _value(4)])
    self.assertEqual(
        filter_cache.filter_values(None, 1, values_iterable),
        [_value(2), _value(3), _value(4)])

    # Check that the cache is used: even if the collection of values is
    # different, the cached results remain unchanged.
    self.assertEqual(
        filter_cache.filter_values(even_filter, 1, None),
        [_value(2), _value(4)])
    self.assertEqual(
        filter_cache.filter_values(None, 1, None),
        [_value(2), _value(3), _value(4)])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for collect_tensor_data."""

from absl.testing import absltest
from absl.testing import parameterized
import mock
import six
import tensorflow as tf
from tf_coder.datasets import collect_tensor_data
from tf_coder.value_search import all_operations
from tf_coder.value_search import value as value_module
from tf_coder.value_search import value_search_settings as settings_module


def new_input(raw_value):
  return value_module.InputValue(raw_value, 'NEW_INPUT')


class CollectTensorDataTest(parameterized.TestCase):

  def setUp(self):
    super(CollectTensorDataTest, self).setUp()
    self.settings = settings_module.default_settings()

    operations = all_operations.get_operations()
    self.unique_with_counts_operation = all_operations.find_operation_with_name(
        'tf.unique_with_counts(x)', operation_list=operations)
    self.indexing_operation = all_operations.find_operation_with_name(
        'IndexingOperation', operation_list=operations)
    self.gather_operation = all_operations.find_operation_with_name(
        'tf.gather(params, indices)', operation_list=operations)
    self.add_operation = all_operations.find_operation_with_name(
        'tf.add(x, y)', operation_list=operations)

    # Example with many operations.
    in1 = value_module.InputValue([1, 1, 2, 5, 6, 5], 'in1')
    in2 = value_module.InputValue([0, 10, 20, 30, 40, 50, 60, 70], 'in2')
    constant_1 = value_module.ConstantValue(1)

    unique = self.unique_with_counts_operation.apply([in1], self.settings)
    indexed = self.indexing_operation.apply([unique, constant_1], self.settings)
    gathered = self.gather_operation.apply([in2, in1], self.settings)
    self.example_value_1 = self.add_operation.apply([indexed, gathered],
                                                    self.settings)

    self.assertEqual(
        self.example_value_1.reconstruct_expression(),
        'tf.add(tf.unique_with_counts(in1)[1], tf.gather(in2, in1))')
    self.assertEqual(self.example_value_1,
                     value_module.OutputValue([10, 10, 21, 52, 63, 52]))

    # Example with many variables and new inputs.
    in3 = value_module.InputValue([1], 'in3')
    in4 = value_module.InputValue([2], 'in4')

    a = self.add_operation.apply([in3, new_input([10])], self.settings)
    b = self.add_operation.apply([in4, in3], self.settings)
    c = self.add_operation.apply([new_input([20]), in3], self.settings)
    d = self.add_operation.apply([a, b], self.settings)
    self.example_value_2 = self.add_operation.apply([c, d], self.settings)

    self.assertEqual(
        self.example_value_2.reconstruct_expression(),
        'tf.add(tf.add(NEW_INPUT, in3), '
        'tf.add(tf.add(in3, NEW_INPUT), tf.add(in4, in3)))')
    self.assertEqual(self.example_value_2, value_module.OutputValue([35]))

  def test_extract_values_with_collapsed_subtrees(self):
    expected_expressions = [
        'tf.add(tf.unique_with_counts(in1)[1], tf.gather(in2, in1))',
        'tf.add(tf.unique_with_counts(in1)[1], NEW_INPUT)',
        'tf.add(NEW_INPUT[1], tf.gather(in2, in1))',
        'tf.add(NEW_INPUT[1], NEW_INPUT)',
        'tf.add(NEW_INPUT, tf.gather(in2, in1))',
        'tf.add(NEW_INPUT, NEW_INPUT)',
        'NEW_INPUT',
    ]
    results = collect_tensor_data.extract_values_with_collapsed_subtrees(
        self.example_value_1)
    # The order doesn't matter except that 'NEW_INPUT' is last.
    self.assertEqual([v.reconstruct_expression() for v in results],
                     expected_expressions)

  def test_count_num_inputs(self):
    self.assertEqual(collect_tensor_data.count_num_inputs(self.example_value_2),
                     (4, {'in3': 3, 'in4': 1, 'NEW_INPUT': 2}))

  def test_normalize_names_and_extract_values(self):
    name_counter = {'in3': 3, 'in4': 1, 'NEW_INPUT': 2}
    copied_value = self.example_value_2.copy()
    name_value_map = collect_tensor_data.normalize_names_and_extract_values(
        copied_value, name_counter)
    self.assertEqual(
        copied_value.reconstruct_expression(use_cache=False),
        # in3 -> in1, in4 -> in2, and NEW_INPUT -> in3 and in4
        'tf.add(tf.add(in3, in1), tf.add(tf.add(in1, in4), tf.add(in2, in1)))')

    self.assertEqual(name_value_map,
                     # Check the raw values.
                     {'in1': value_module.OutputValue([1]),
                      'in2': value_module.OutputValue([2]),
                      'in3': value_module.OutputValue([20]),
                      'in4': value_module.OutputValue([10])})
    for name in name_value_map:
      self.assertIsInstance(name_value_map[name], value_module.InputValue)
      self.assertEqual(name, name_value_map[name].name)

  def test_extract_operations(self):
    self.assertCountEqual(
        collect_tensor_data.extract_operations(self.example_value_1),
        [self.unique_with_counts_operation,
         self.indexing_operation,
         self.gather_operation,
         self.add_operation])
    self.assertCountEqual(
        collect_tensor_data.extract_operations(self.example_value_2),
        [self.add_operation] * 5)

  @parameterized.named_parameters(
      ('1', 1, []),
      ('2', 2, ['tf.add(tf.unique_with_counts(in1)[1], tf.gather(in2, in1))',
                'tf.add(tf.unique_with_counts(in1)[1], in2)',
                'tf.add(in1[1], in2)',
                'tf.add(in1, in2)']),
      ('3', 3, ['tf.add(tf.unique_with_counts(in1)[1], tf.gather(in2, in1))',
                'tf.add(tf.unique_with_counts(in1)[1], in2)',
                'tf.add(in3[1], tf.gather(in2, in1))',
                'tf.add(in1[1], in2)',
                'tf.add(in3, tf.gather(in2, in1))',
                'tf.add(in1, in2)']))
  def test_extract_examples_from_value(self, max_num_inputs,
                                       expected_expressions):
    actual = collect_tensor_data.extract_examples_from_value(
        self.example_value_1, max_num_inputs=max_num_inputs)

    # Check that all expressions are as expected.
    self.assertCountEqual([example.expression for example in actual],
                          expected_expressions)

    # Check all elements of one IOExample namedtuple. This example is at index 1
    # when max_num_inputs > 1.
    if max_num_inputs > 1:
      expected_index_1 = collect_tensor_data.IOExample(
          expression='tf.add(tf.unique_with_counts(in1)[1], in2)',
          input_values=[value_module.InputValue([1, 1, 2, 5, 6, 5], 'in1'),
                        value_module.InputValue([10, 10, 20, 50, 60, 50],
                                                'in2')],
          output_value=value_module.OutputValue([10, 10, 21, 52, 63, 52]),
          num_inputs=2,
          operations=[self.add_operation,
                      self.indexing_operation,
                      self.unique_with_counts_operation])
      self.assertEqual(actual[1], expected_index_1)
      # Equality of Value objects is done by comparing the wrapped values. Check
      # the names in input_values too.
      for actual_value, expected_value in zip(actual[1].input_values,
                                              expected_index_1.input_values):
        self.assertIsInstance(actual_value, value_module.InputValue)
        self.assertEqual(actual_value.name, expected_value.name)

    # Check that all extracted examples actually work by eval-ing them.
    for example in actual:
      namespace_dict = {'tf': tf}
      self.assertLen(example.input_values, example.num_inputs)
      for input_value in example.input_values:
        namespace_dict[input_value.name] = input_value.value
      eval_output = eval(example.expression, namespace_dict)  # pylint: disable=eval-used
      self.assertEqual(value_module.OutputValue(eval_output),
                       example.output_value)
      self.assertEqual(example.output_value, self.example_value_1)

  def test_extract_examples_from_value_without_inputs(self):
    constant_1 = value_module.ConstantValue(1)
    constant_2 = value_module.ConstantValue(2)
    constant_3 = value_module.ConstantValue(3)

    subtree = self.add_operation.apply([constant_1, constant_2], self.settings)
    without_inputs = self.add_operation.apply([subtree, constant_3],
                                              self.settings)
    actual = collect_tensor_data.extract_examples_from_value(without_inputs)

    self.assertCountEqual(
        [example.expression for example in actual],
        # `tf.add(tf.add(1, 2), 3)` has no inputs and is not included.
        ['tf.add(in1, 3)'])

  def test_create_tf_examples(self):
    sparse_tensor = tf.SparseTensor(
        values=[0, -15, 30],
        indices=[[12], [34], [56]],
        dense_shape=[100])
    # This example does not represent a realistic tensor transformation. It uses
    # variety in the input/output tensors to exercise the featurization.
    io_example = collect_tensor_data.IOExample(
        expression='tf.dummy_expression(in1, in2)',
        input_values=[
            value_module.InputValue([[[0.5, 2.5, 9.0],
                                      [-0.25, 0.0, 1.25]]], 'in1'),
            value_module.InputValue(sparse_tensor, 'in2'),
        ],
        output_value=value_module.OutputValue([[1.0], [0.0], [1.0], [0.0]]),
        num_inputs=2,
        operations=[self.add_operation, self.add_operation,
                    self.gather_operation])

    with mock.patch.object(collect_tensor_data, 'COUNT_BOUNDARIES',
                           new=[0, 1, 3, 50, float('inf')]):
      with mock.patch.object(
          collect_tensor_data, 'FLOAT_BOUNDARIES',
          new=[-float('inf'), -10, -1e-8, 1e-8, 10, float('inf')]):
        tf_examples = collect_tensor_data.create_tf_examples(io_example)

    operation_list = all_operations.get_operations(
        include_sparse_operations=True)
    expected_operations = [
        2 if op.name == 'tf.add(x, y)'
        else 1 if op.name == 'tf.gather(params, indices)' else 0
        for op in operation_list]

    expected_tf_example_1 = {
        # Features from featurize_value.
        'kind': [2, 2, 3, 0],
        'dtype': [8, 8, 0, 0],
        'rank': [2, 3, 1, 0],
        'shape': [4, 1, 0, 0,
                  1, 2, 3, 0,
                  100, 0, 0, 0,
                  0, 0, 0, 0],
        'shape_buckets': [2, 1, 0, 0,
                          1, 1, 2, 0,
                          3, 0, 0, 0,
                          0, 0, 0, 0],
        'floats': [1.0, 0.0, 0.5, 0.5,
                   9.0, -0.25, 13/6, 13.5/6,
                   30, -15, 5, 15,
                   0, 0, 0, 0],
        'float_buckets': [3, 2, 3, 3,
                          3, 1, 3, 3,
                          4, 0, 3, 4,
                          2, 2, 2, 2],
        'counts': [4, 4, 2, 2, 2, 0, 4, 2,
                   6, 6, 4, 1, 0, 1, 2, 6,
                   100, 3, 1, 1, 0, 1, 1, 3,
                   1, 1, 0, 1, 0, 0, 1, 1],
        'count_buckets': [2, 2, 1, 1, 1, 0, 2, 1,
                          2, 2, 2, 1, 0, 1, 1, 2,
                          3, 2, 1, 1, 0, 1, 1, 2,
                          1, 1, 0, 1, 0, 0, 1, 1],
        'fractions': [4/4, 2/4, 2/4, 2/4, 0/4, 4/4, 2/4,
                      6/6, 4/6, 1/6, 0/6, 1/6, 2/6, 6/6,
                      3/100, 1/3, 1/3, 0/3, 1/3, 1/3, 3/3,
                      1/1, 0/1, 1/1, 0/1, 0/1, 1/1, 1/1],
        'booleans': [1, 1, 0, 1, 0, 1, 1, 0,
                     1, 1, 0, 0, 0, 0, 0, 1,
                     0, 1, 0, 0, 0, 0, 0, 1,
                     1, 1, 0, 1, 0, 1, 1, 1],
        'value_string': [
            b'tf.float32:[[1.0], [0.0], [1.0], [0.0]]',
            b'tf.float32:[[[0.5, 2.5, 9.0], [-0.25, 0.0, 1.25]]]',
            (b'SparseTensor(indices=tf.Tensor(\n[[12]\n [34]\n [56]], '
             b'shape=(3, 1), dtype=int64), '
             b'values=tf.Tensor([  0 -15  30], shape=(3,), dtype=int32), '
             b'dense_shape=tf.Tensor([100], shape=(1,), dtype=int64))'),
            b'0'],

        # Features from featurize_input_and_output.
        'io_comparisons': [2, 2, 2, 0, 2, 2, 1,
                           2, 0, 0, 2, 0, 1, 1,
                           1, 1, 1, 1, 1, 1, 1],
        'io_booleans': [0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0,
                        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                        1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
        'io_counts': [1, 2, 1,
                      1, 2, 1,
                      1, 1, 1],
        'io_count_buckets': [1, 1, 1,
                             1, 1, 1,
                             1, 1, 1],
        'io_fractions': [1/6, 2/4, 1/6, 1/4,
                         1/3, 2/4, 1/3, 1/4,
                         1/1, 1/1, 1/1, 1/1],

        # Features added in create_examples.
        'num_inputs': [2],
        'operations': expected_operations,
        'expression': [b'tf.dummy_expression(in1, in2)'],
    }

    print(tf_examples)

    self.assertLen(tf_examples, 2)
    actual_tf_example_1, actual_tf_example_2 = tf_examples

    # Check the entire first example.
    for key, expected in six.iteritems(expected_tf_example_1):
      some_list = actual_tf_example_1.features.feature[key]
      if some_list.HasField('float_list'):
        actual = some_list.float_list.value
        actual = [round(f, 6) for f in actual]
        expected = [round(f, 6) for f in expected]
      elif some_list.HasField('int64_list'):
        actual = some_list.int64_list.value
      elif some_list.HasField('bytes_list'):
        actual = some_list.bytes_list.value
      else:
        self.fail('Failed to extract list from TF example.')

      # Printing side-by-side like this in the test log is more helpful than the
      # AssertionError message. Look at the Python3 log, which prints ints
      # without the L suffix.
      print('key: {}\n'
            '  expected: {}\n'
            '  got:      {}'.format(key, expected, actual))
      self.assertEqual(actual, expected)

    # Less rigorous checks for the second example, where the two inputs have
    # swapped.
    self.assertEqual(
        actual_tf_example_2.features.feature['rank'].int64_list.value,
        [2, 1, 3, 0])
    self.assertEqual(
        actual_tf_example_2.features.feature['shape'].int64_list.value,
        [4, 1, 0, 0,
         100, 0, 0, 0,
         1, 2, 3, 0,
         0, 0, 0, 0])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for description_handler_factory.py."""

from absl.testing import absltest
from tf_coder.natural_language import description_handler_factory


class DescriptionHandlerFactoryTest(absltest.TestCase):

  def test_handler_string_list(self):
    string_list = description_handler_factory.handler_string_list()
    self.assertIn('no_change', string_list)

  def test_create_handler(self):
    handler = description_handler_factory.create_handler('no_change')
    self.assertEqual(repr(handler), 'NoChangeDescriptionHandler')

  def test_create_handler_raises(self):
    with self.assertRaises(ValueError):
      description_handler_factory.create_handler('nonexistent_handler')

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""An interface for handling natural language descriptions for TF-Coder."""

import abc
import re
from typing import Dict, List, Optional, Text

import six
from tf_coder.benchmarks import benchmark as benchmark_module
from tf_coder.value_search import all_operations
from tf_coder.value_search import operation_base
from tf_coder.value_search import value_search_settings as settings_module


@six.add_metaclass(abc.ABCMeta)
class DescriptionHandler(object):
  """Handles a natural language description of a task.

  Attributes:
    operations: A list of operations that the handler knows about.
    all_names: A list of operation names, in the same order as the `operations`
      list.
  """

  def __init__(self,
               operations: Optional[List[operation_base.Operation]] = None):
    """Initializes the handler.

    Args:
      operations: A list of operations that the scorer should handle. Exposed
        for testing.
    Raises:
      ValueError: If there are duplicate operation names.
    """
    self.operations = (
        operations if operations
        else all_operations.get_operations(include_sparse_operations=True))
    self.all_names = [operation.name for operation in self.operations]
    if len(set(self.all_names)) != len(self.operations):
      raise ValueError('Duplicate operation name.')

  @abc.abstractmethod
  def get_operation_multipliers(
      self, benchmark: benchmark_module.Benchmark,
      settings: settings_module.Settings) -> Dict[Text, float]:
    """Returns a map from operation names to their weight multiplier.

    The weight multiplier should be between 0 and 1 if the operation should be
    prioritized, or greater than 1 if it should be deprioritized.

    Args:
      benchmark: Benchmark object corresponding to the TF-Coder task.
      settings: A Settings object storing settings for this search.

    Returns:
      A map from operation name to weight multiplier, such that the operation
      with that name should have its weight modified by that multiplier. If the
      dict does not contain a key, it means the weight should not be modified
      (equivalent to a multiplier of 1).
    """

  def __repr__(self) -> Text:
    """Returns a string containing details about this handler and parameters."""
    return self.__class__.__name__


class NoChangeDescriptionHandler(DescriptionHandler):
  """A description handler that does not change any operation weights."""

  def get_operation_multipliers(
      self, benchmark: benchmark_module.Benchmark,
      settings: settings_module.Settings) -> Dict[Text, float]:
    """See base class."""
    return {}


class FunctionNameDescriptionHandler(DescriptionHandler):
  """Prioritizes functions with names that appear in the docstring."""

  def __init__(self,
               operations: Optional[List[operation_base.Operation]] = None,
               multiplier: float = 0.75):
    """Creates a FunctionNameDescriptionHandler.

    Args:
      operations: A list of operations that the scorer should handle. Exposed
        for testing.
      multiplier: The multiplier applied to an operation's weight if it is
        prioritized.
    """
    super(FunctionNameDescriptionHandler, self).__init__(operations)
    self.multiplier = multiplier

  def get_operation_multipliers(
      self, benchmark: benchmark_module.Benchmark,
      settings: settings_module.Settings) -> Dict[Text, float]:
    """See base class."""
    description = benchmark.description.lower()
    multipliers = {}
    for name in self.all_names:
      if name.startswith('tf.') and '(' in name:
        function_name = name[len('tf.') : name.index('(')].lower()
        function_name_parts = re.split(r'[._]', function_name)
        if all(part in description for part in function_name_parts):
          if settings.printing.prioritized_operations:
            print('FunctionNameDescriptionHandler prioritized {}'.format(name))
          multipliers[name] = self.multiplier
    return multipliers

  def __repr__(self) -> Text:
    """See base class."""
    return '{}(multiplier={})'.format(self.__class__.__name__, self.multiplier)



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A script for using TF-Coder (an alternative to using the Colab notebook).

Usage:
1. Edit `get_problem()` to specify your problem.
2. If desired, edit `get_settings()` to specify settings for TF-Coder.
3. Run this file, e.g., `python3 tf_coder_main.py`.
"""

import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # Must happen before importing tf.
os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # CPU is faster than GPU.

from absl import app  # pylint: disable=g-import-not-at-top
import numpy as np  # pylint: disable=unused-import
import tensorflow as tf  # pylint: disable=unused-import

from tf_coder.value_search import colab_interface
from tf_coder.value_search import value_search_settings as settings_module


def get_problem():
  """Specifies a problem to run TF-Coder on. Edit this function!"""
  # A dict mapping input variable names to input tensors.
  inputs = {
      'rows': [10, 20, 30],
      'cols': [1, 2, 3, 4],
  }

  # The single desired output tensor.
  output = [[11, 12, 13, 14],
            [21, 22, 23, 24],
            [31, 32, 33, 34]]

  # A list of relevant scalar constants (if any).
  constants = []

  # An English description of the tensor manipulation.
  description = 'add two vectors with broadcasting to get a matrix'

  return inputs, output, constants, description


def get_settings():
  """Specifies settings for TF-Coder. Edit this function!"""
  # How long to search for a solution, in seconds.
  time_limit = 300

  # How many solutions to find before stopping. If more than 1, the entire
  # search will slow down.
  number_of_solutions = 1

  # Whether solutions must use all inputs, at least one input, or no such
  # requirement. Choose one of "all inputs", "one input", "no restriction".
  solution_requirement = 'all inputs'
  assert solution_requirement in ['all inputs', 'one input', 'no restriction']

  return settings_module.from_dict({
      'timeout': time_limit,
      'only_minimal_solutions': False,
      'max_solutions': number_of_solutions,
      'require_all_inputs_used': solution_requirement == 'all inputs',
      'require_one_input_used': solution_requirement == 'one input',
  })


def run_tf_coder(inputs, output, constants, description, settings):
  """Runs TF-Coder on a problem, using the given settings."""
  # Results will be printed to standard output.
  colab_interface.run_value_search_from_colab(
      inputs, output, constants, description, settings)


def print_supported_operations():
  """Run this function to print all supported operations."""
  colab_interface.print_supported_operations()


def main(unused_argv):
  # It takes several seconds to load the models.
  colab_interface.warm_up()

  inputs, output, constants, description = get_problem()
  settings = get_settings()

  run_tf_coder(inputs, output, constants, description, settings)


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for python_operations.py."""

from absl import logging
from absl.testing import absltest
from absl.testing import parameterized
import tensorflow as tf
from tf_coder.value_search import python_operations
from tf_coder.value_search import value
from tf_coder.value_search import value_search_settings as settings_module


class IndexingOperationTest(parameterized.TestCase):

  def setUp(self):
    super(IndexingOperationTest, self).setUp()
    self.operation = python_operations.IndexingOperation()

  @parameterized.named_parameters(
      ('index_0', 0, 1.2),
      ('index_2', 2, 5.6),
      ('index_negative_2', -2, 3.4))
  def test_apply_for_list(self, index, expected):
    arg_values = [value.ConstantValue([1.2, 3.4, 5.6]),
                  value.ConstantValue(index)]
    result_value = self.operation.apply(arg_values,
                                        settings_module.default_settings())
    self.assertEqual(result_value, value.ConstantValue(expected))

  @parameterized.named_parameters(
      ('index_0', 0, [1.2, 3.4]),
      ('index_negative_1', -1, [5.6, 7.8]))
  def test_apply_for_tensor(self, index, expected):
    tensor = tf.constant([[1.2, 3.4], [5.6, 7.8]])
    arg_values = [value.ConstantValue(tensor),
                  value.ConstantValue(index)]
    result_value = self.operation.apply(arg_values,
                                        settings_module.default_settings())
    self.assertEqual(result_value, value.ConstantValue(tf.constant(expected)))

  def test_reconstruct_expression(self):
    to_index = value.InputValue([1.2, 3.4], 'my_list')
    index = value.ConstantValue(0)
    self.assertEqual(self.operation.reconstruct_expression([to_index, index]),
                     'my_list[0]')


class IndexingAxis1OperationTest(absltest.TestCase):

  def setUp(self):
    super(IndexingAxis1OperationTest, self).setUp()
    self.operation = python_operations.IndexingAxis1Operation()
    self.arg_values = [value.InputValue([[12, 34], [56, 78]], 'my_input'),
                       value.ConstantValue(1)]

  def test_apply(self):
    self.assertEqual(self.operation.apply(self.arg_values,
                                          settings_module.default_settings()),
                     value.OutputValue([34, 78]))

  def test_reconstruct_expression(self):
    self.assertEqual(self.operation.reconstruct_expression(self.arg_values),
                     'my_input[:, 1]')


class SingletonTupleCreationOperationTest(absltest.TestCase):

  def setUp(self):
    super(SingletonTupleCreationOperationTest, self).setUp()
    self.operation = python_operations.SingletonTupleCreationOperation()
    self.arg_values = [value.ConstantValue(12)]

  def test_apply(self):
    self.assertEqual(self.operation.apply(self.arg_values,
                                          settings_module.default_settings()),
                     value.ConstantValue((12,)))

  def test_reconstruct_expression(self):
    self.assertEqual(self.operation.reconstruct_expression(self.arg_values),
                     '(12,)')


class PairCreationOperationTest(absltest.TestCase):

  def setUp(self):
    super(PairCreationOperationTest, self).setUp()
    self.operation = python_operations.PairCreationOperation()
    self.arg_values = [value.ConstantValue(12), value.ConstantValue(34)]

  def test_apply(self):
    self.assertEqual(self.operation.apply(self.arg_values,
                                          settings_module.default_settings()),
                     value.ConstantValue((12, 34)))

  def test_reconstruct_expression(self):
    self.assertEqual(self.operation.reconstruct_expression(self.arg_values),
                     '(12, 34)')


class TripleCreationOperationTest(absltest.TestCase):

  def setUp(self):
    super(TripleCreationOperationTest, self).setUp()
    self.operation = python_operations.TripleCreationOperation()
    self.arg_values = [value.ConstantValue(12),
                       value.ConstantValue(34),
                       value.ConstantValue(56)]

  def test_apply(self):
    self.assertEqual(self.operation.apply(self.arg_values,
                                          settings_module.default_settings()),
                     value.ConstantValue((12, 34, 56)))

  def test_reconstruct_expression(self):
    self.assertEqual(self.operation.reconstruct_expression(self.arg_values),
                     '(12, 34, 56)')


class SlicingAxis0LeftOperationTest(absltest.TestCase):

  def setUp(self):
    super(SlicingAxis0LeftOperationTest, self).setUp()
    self.operation = python_operations.SlicingAxis0LeftOperation()
    self.arg_values = [value.InputValue([12, 34, 56], 'my_input'),
                       value.ConstantValue(1)]

  def test_get_docstring(self):
    self.assertIn('Selects a suffix of indices along axis 0 of the tensor',
                  self.operation.metadata.docstring)

  def test_apply(self):
    self.assertEqual(self.operation.apply(self.arg_values,
                                          settings_module.default_settings()),
                     value.OutputValue([34, 56]))

  def test_reconstruct_expression(self):
    self.assertEqual(self.operation.reconstruct_expression(self.arg_values),
                     'my_input[1:]')


class SlicingAxis0RightOperationTest(absltest.TestCase):

  def setUp(self):
    super(SlicingAxis0RightOperationTest, self).setUp()
    self.operation = python_operations.SlicingAxis0RightOperation()
    self.arg_values = [value.InputValue([12, 34, 56], 'my_input'),
                       value.ConstantValue(-1)]

  def test_get_docstring(self):
    self.assertIn('Selects a prefix of indices along axis 0 of the tensor',
                  self.operation.metadata.docstring)

  def test_apply(self):
    self.assertEqual(self.operation.apply(self.arg_values,
                                          settings_module.default_settings()),
                     value.OutputValue([12, 34]))

  def test_reconstruct_expression(self):
    self.assertEqual(self.operation.reconstruct_expression(self.arg_values),
                     'my_input[:-1]')


class SlicingAxis0BothOperationTest(absltest.TestCase):

  def setUp(self):
    super(SlicingAxis0BothOperationTest, self).setUp()
    self.operation = python_operations.SlicingAxis0BothOperation()
    self.arg_values = [value.InputValue([12, 34, 56, 78], 'my_input'),
                       value.ConstantValue(1),
                       value.ConstantValue(-1)]

  def test_get_docstring(self):
    self.assertIn('Selects a range of indices along axis 0 of the tensor',
                  self.operation.metadata.docstring)

  def test_apply(self):
    self.assertEqual(self.operation.apply(self.arg_values,
                                          settings_module.default_settings()),
                     value.OutputValue([34, 56]))

  def test_reconstruct_expression(self):
    self.assertEqual(self.operation.reconstruct_expression(self.arg_values),
                     'my_input[1:-1]')


class SlicingAxis1LeftOperationTest(absltest.TestCase):

  def setUp(self):
    super(SlicingAxis1LeftOperationTest, self).setUp()
    self.operation = python_operations.SlicingAxis1LeftOperation()
    self.arg_values = [value.InputValue([[12, 34, 56]], 'my_input'),
                       value.ConstantValue(1)]

  def test_get_docstring(self):
    self.assertIn('Selects a suffix of indices along axis 1 of the tensor',
                  self.operation.metadata.docstring)

  def test_apply(self):
    self.assertEqual(self.operation.apply(self.arg_values,
                                          settings_module.default_settings()),
                     value.OutputValue([[34, 56]]))

  def test_reconstruct_expression(self):
    self.assertEqual(self.operation.reconstruct_expression(self.arg_values),
                     'my_input[:, 1:]')


class SlicingAxis1RightOperationTest(absltest.TestCase):

  def setUp(self):
    super(SlicingAxis1RightOperationTest, self).setUp()
    self.operation = python_operations.SlicingAxis1RightOperation()
    self.arg_values = [value.InputValue([[12, 34, 56]], 'my_input'),
                       value.ConstantValue(-1)]

  def test_get_docstring(self):
    self.assertIn('Selects a prefix of indices along axis 1 of the tensor',
                  self.operation.metadata.docstring)

  def test_apply(self):
    self.assertEqual(self.operation.apply(self.arg_values,
                                          settings_module.default_settings()),
                     value.OutputValue([[12, 34]]))

  def test_reconstruct_expression(self):
    self.assertEqual(self.operation.reconstruct_expression(self.arg_values),
                     'my_input[:, :-1]')


class SlicingAxis1BothOperationTest(absltest.TestCase):

  def setUp(self):
    super(SlicingAxis1BothOperationTest, self).setUp()
    self.operation = python_operations.SlicingAxis1BothOperation()
    self.arg_values = [value.InputValue([[12, 34, 56, 78],
                                         [-1, -2, -3, -4]], 'my_input'),
                       value.ConstantValue(1),
                       value.ConstantValue(-1)]

  def test_get_docstring(self):
    self.assertIn('Selects a range of indices along axis 1 of the tensor',
                  self.operation.metadata.docstring)

  def test_apply(self):
    self.assertEqual(self.operation.apply(self.arg_values,
                                          settings_module.default_settings()),
                     value.OutputValue([[34, 56], [-2, -3]]))

  def test_reconstruct_expression(self):
    self.assertEqual(self.operation.reconstruct_expression(self.arg_values),
                     'my_input[:, 1:-1]')


if __name__ == '__main__':
  logging.set_verbosity(logging.ERROR)

  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Defines the OperationStatistics class.

An OperationStatistics instance tracks how often operations are applied, how
many of those applications are successful, and how long they take.
"""

import collections

# Format strings for the statistics table.
_HEADER_FORMAT_STR = ('{name:64}{eps:>12}{sps:>12}{executions:>13}'
                      '{successes:>12}{rate:>15}{time:>13}{time_frac:>11}')
_ROW_FORMAT_STR = ('{name:64}{eps:12.1f}{sps:12.1f}{executions:13}'
                   '{successes:12}{rate:15.2%}{time:13.3f}{time_frac:11.1%}')


def _nan_div(numerator, denominator):
  """Performs division, returning NaN if the denominator is zero."""
  return numerator / denominator if denominator != 0 else float('NaN')


class OperationStatistics(object):
  """Tracks statistics about applications of Operations.

  An OperationStatistics instance tracks how often Operation objects are
  applied, how many applications are successful, and how long they take.

  Attributes:
    total_apply_count: The total number of Operation applications.
    total_apply_successes: The total number of successful Operation
      applications.
    operation_apply_time: A dict mapping Operation names to the total time spent
      applying that Operation.
    operation_apply_count: A dict mapping Operation names to the number of
      applications of that Operation.
    operation_apply_successes: A dict mapping Operation names to the number of
      successful applications of that Operation.
    all_operation_names: A set of recorded Operation names.
  """

  def __init__(self):
    """Initializes the attributes."""
    self.total_apply_count = 0
    self.total_apply_successes = 0
    self.operation_apply_time = collections.defaultdict(float)
    self.operation_apply_count = collections.Counter()
    self.operation_apply_successes = collections.Counter()
    self.all_operation_names = set()

  def update(self, operation_name, count, successes, time):
    """Updates the statistics with the given statistics for one operation."""
    self.total_apply_count += count
    self.total_apply_successes += successes
    self.operation_apply_time[operation_name] += time
    self.operation_apply_count[operation_name] += count
    self.operation_apply_successes[operation_name] += successes
    self.all_operation_names.add(operation_name)

  def get_total_time(self):
    """Returns the total time spent applying operations."""
    return sum(self.operation_apply_time.values())

  def statistics_as_string(self, operation_names=None, num_unique_values=None,
                           elapsed_time=None, sort_by_time=True):
    """Returns a printable string with statistics for the given operations.

    Args:
      operation_names: A list of operation names. If None, all recorded
        operations will be logged.
      num_unique_values: The number of unique values found by value search. This
        is optional; if provided, it will be logged.
      elapsed_time: The total time used by value search (not just for Operation
        applications). This is optional; if provided, the operation applications
        per second will be logged.
      sort_by_time: Whether to sort the table by decreasing time. If False
        (default), then the table will be sorted by name alphabetically.
    """
    string_parts = []

    if operation_names is None:
      operation_names = list(self.all_operation_names)

    header = _HEADER_FORMAT_STR.format(
        name='Operation name', eps='Exec./sec', sps='Succ./sec',
        executions='Executions', successes='Successes', rate='Success rate',
        time='Time (sec)', time_frac='Time (%)')
    string_parts.append(header)
    string_parts.append('-' * len(header))
    total_time = sum(self.operation_apply_time[name]
                     for name in operation_names)
    if sort_by_time:
      sorted_names = sorted(operation_names,
                            key=lambda n: -self.operation_apply_time[n])
    else:
      sorted_names = sorted(operation_names)
    for name in sorted_names:
      apply_time = self.operation_apply_time[name]
      apply_count = self.operation_apply_count[name]
      apply_successes = self.operation_apply_successes[name]
      string_parts.append(_ROW_FORMAT_STR.format(
          name=name,
          eps=_nan_div(apply_count, apply_time),
          sps=_nan_div(apply_successes, apply_time),
          executions=apply_count,
          successes=apply_successes,
          rate=_nan_div(apply_successes, apply_count),
          time=apply_time,
          time_frac=_nan_div(apply_time, total_time)))

    string_parts.append('\nNumber of evaluations: {}\n'
                        'Number of successful evaluations: {}\n'
                        'Total time applying operations: {:.2f} sec\n'.format(
                            self.total_apply_count,
                            self.total_apply_successes,
                            self.get_total_time()))
    if num_unique_values is not None:
      string_parts.append('Number of unique values: {}'.format(
          num_unique_values))
    if elapsed_time is not None:
      string_parts.append('Executions per second: {:.1f}'.format(
          _nan_div(self.total_apply_count, elapsed_time)))

    return '\n'.join(string_parts)



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for settings_module.py."""

import json

from absl.testing import absltest
from absl.testing import parameterized
from tf_coder.value_search import value_search_settings as settings_module


class ValueSearchSettingsTest(parameterized.TestCase):

  def test_set(self):
    settings = settings_module.Settings()
    settings.set('timeout', 12345)
    settings.set('printing.verbose', True)
    settings.set('tensor_model_prioritize_threshold', 0.9)
    self.assertEqual(settings.timeout, 12345)
    self.assertTrue(settings.printing.verbose)
    self.assertEqual(settings.tensor_model.prioritize_threshold, 0.9)

  def test_set_raises(self):
    settings = settings_module.Settings()
    with self.assertRaises(ValueError):
      settings.set('unknown_setting', True)
    with self.assertRaises(ValueError):
      settings.set('printing.unknown_setting', True)

  def test_from_dict(self):
    overrides = {
        'timeout': 12345,
        'printing.verbose': True,
        'tensor_model_prioritize_threshold': 0.9,
    }
    settings = settings_module.from_dict(overrides)
    self.assertEqual(settings.timeout, 12345)
    self.assertTrue(settings.printing.verbose)
    self.assertEqual(settings.tensor_model.prioritize_threshold, 0.9)

  def test_from_list(self):
    overrides = [
        'timeout=12345',
        'printing.verbose=True',
        'tensor_model_prioritize_threshold=0.9',
    ]
    settings = settings_module.from_list(overrides)
    self.assertEqual(settings.timeout, 12345)
    self.assertTrue(settings.printing.verbose)
    self.assertEqual(settings.tensor_model.prioritize_threshold, 0.9)

  @parameterized.named_parameters(
      ('no_equals', 'timeout: 12345'),
      ('too_many_equals', 'timeout==12345'),
      ('spaces', 'timeout = 12345'),
      ('bad_name', 'unknown_setting=12345'),
      ('bad_value', 'timeout=foo()'))
  def test_from_list_raises(self, override_string):
    with self.assertRaises(ValueError):
      settings_module.from_list([override_string])

  def test_as_dict_default_works_with_json(self):
    settings = settings_module.default_settings()
    as_dict = settings.as_dict()
    self.assertFalse(as_dict['printing.verbose'])
    # Test that all defaults are JSON-serializable.
    self.assertNotEmpty(json.dumps(as_dict))


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Exhaustive value search (enumerating by weight of expression)."""

import gc
import json
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # Must happen before importing tf.
os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # CPU is faster than GPU.
import sys  # pylint: disable=g-import-not-at-top

from absl import app
from absl import flags
import numpy as np
from scipy.stats import mstats
from tf_coder.benchmarks import all_benchmarks
from tf_coder.benchmarks import autopandas_benchmarks
from tf_coder.benchmarks import google_benchmarks
from tf_coder.benchmarks import stackoverflow_benchmarks
from tf_coder.datasets import collect_tensor_data
from tf_coder.models import tensor_features_model
from tf_coder.natural_language import description_handler_factory
from tf_coder.value_search import value_search
from tf_coder.value_search import value_search_settings as settings_module


FLAGS = flags.FLAGS

flags.DEFINE_string('benchmark_name', 'ALL',
                    'The name of a benchmark to run, or "ALL".')
flags.DEFINE_multi_string('settings',
                          ['timeout=300'],
                          'Settings to override the defaults.')
flags.DEFINE_string(
    'json_output',
    '',
    'Where the results json file should be written. It will only be written if '
    'all benchmarks are run. Use an empty string to avoid writing this file.')
flags.DEFINE_string(
    'notes',
    '',
    'Any notes to write into the results json file.')


def benchmark_name_validator(benchmark_name):
  """Checks that benchmark_name is "ALL", or refers to exactly one benchmark."""
  return (benchmark_name == 'ALL' or
          all_benchmarks.find_benchmark_with_name(benchmark_name) is not None)

flags.register_validator('benchmark_name', benchmark_name_validator,
                         message=('benchmark_name must be "ALL" or refer to '
                                  'exactly one benchmark.'))




def run_on_all_benchmarks(settings, description_handler, json_output,
                          benchmark_name, notes):
  """Runs value search on all benchmarks, printing results to stdout."""

  benchmark_count = 0
  benchmark_success = 0
  unsolved_benchmarks = []
  solution_times = []  # Only including successful tasks.

  results_json = {
      'benchmark_name': benchmark_name,
      'settings': settings.as_dict(),
      'notes': notes,
      'results': [],
  }

  if (settings.tensor_model.config_path and
      settings.tensor_model.checkpoint_path):
    tensor_config = tensor_features_model.load_config(
        settings.tensor_model.config_path)
    tensor_model = tensor_features_model.get_model(tensor_config)
    checkpoint = tensor_features_model.create_checkpoint(tensor_model)
    checkpoint.restore(settings.tensor_model.checkpoint_path).expect_partial()

    # Warm up. Running the model for the first time takes an extra ~10 seconds.
    print('Warming up the tensor features model...')
    value_search.operation_multipliers_from_tensor_model(
        all_benchmarks.find_benchmark_with_name('simple_cast'),
        tensor_model, tensor_config, settings)
    print('Finished warming up.')
  else:
    tensor_config = None
    tensor_model = None

  print('=' * 80)

  if benchmark_name == 'ALL':
    # Only run on benchmarks from these important modules.
    modules = [google_benchmarks, stackoverflow_benchmarks,
               autopandas_benchmarks]
  else:
    # Allow searching by name among even more benchmark modules.
    modules = None

  for benchmark in all_benchmarks.get_chosen_benchmarks(
      benchmark_name, modules=modules):
    gc.collect()

    print('Performing value search for benchmark {}.\n'
          .format(benchmark.name))
    benchmark_count += 1

    result = value_search.run_value_search(
        benchmark=benchmark,
        settings=settings,
        description_handler=description_handler,
        tensor_model=tensor_model,
        tensor_config=tensor_config)

    if settings.printing.statistics:
      print('\nOperation statistics:\n{}'.format(
          result.statistics.statistics_as_string(
              num_unique_values=len(result.value_set),
              elapsed_time=result.total_time,
              sort_by_time=settings.printing.statistics_sort_by_time)))

    solutions = result.solutions
    if solutions:
      first_solution = solutions[0]
      print('\nBest solution of weight {} found in {:.2f} sec:\n{}'.format(
          first_solution.weight, first_solution.time,
          first_solution.expression))
      benchmark_success += 1
      solution_times.append(first_solution.time)
    else:
      unsolved_benchmarks.append(benchmark)
    print('=' * 80)
    sys.stdout.flush()

    results_json['results'].append({
        'name': benchmark.name,
        'solved': bool(solutions),
        'solution': solutions[0].expression if solutions else None,
        'solution_weight': solutions[0].weight if solutions else None,
        'time': solutions[0].time if solutions else None,
        'operations': (
            [op.name for op in collect_tensor_data.extract_operations(
                solutions[0].value)]
            if solutions else None),
    })

  solve_time_total = sum(solution_times)
  solve_time_mean = np.mean(solution_times)
  solve_time_geometric_mean = mstats.gmean(solution_times)

  results_json['num_benchmarks'] = benchmark_count
  results_json['num_solved'] = benchmark_success
  results_json['solve_time_total'] = solve_time_total
  results_json['solve_time_mean'] = solve_time_mean
  results_json['solve_time_geometric_mean'] = solve_time_geometric_mean

  print('Solved {} out of {} benchmarks in {:.2f} sec.'.format(
      benchmark_success, benchmark_count, solve_time_total))
  print('\n'
        'Arithmetic mean of solve times: {:.2f} sec\n'
        'Geometric mean of solve times: {:.2f} sec\n'.format(
            solve_time_mean, solve_time_geometric_mean))

  print('Unsolved benchmarks:')
  for unsolved in unsolved_benchmarks:
    print('Name: {}, target program: {}'.format(
        unsolved.name, unsolved.target_program))
  print()

  if json_output and benchmark_name == 'ALL':
    with open(json_output, 'w') as json_file:
      json.dump(results_json, json_file,
                indent=4, sort_keys=True, separators=(',', ': '))
      json_file.write('\n')
    print('Wrote JSON results to {}.'.format(json_output))
  else:
    print('Did not write JSON results file.')


def main(unused_argv):

  settings = settings_module.from_list(FLAGS.settings)
  description_handler = description_handler_factory.create_handler(
      settings.description_handler_name)
  print('Description handler: {!r}\n'.format(description_handler))

  run_on_all_benchmarks(settings, description_handler, FLAGS.json_output,
                        FLAGS.benchmark_name, FLAGS.notes)


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Benchmarks adapted from AutoPandas's benchmarks.

Autopandas benchmarks are at:
https://github.com/rbavishi/autopandas/blob/master/autopandas_v2/evaluation/benchmarks/stackoverflow.py
"""

# Avoid wrapping URLs and target programs to ease clicking and copying.
# pylint: disable=line-too-long

# Every function in this module takes no arguments and creates a benchmark.
# pylint: disable=missing-docstring,g-doc-return-or-yield

from tf_coder.benchmarks import benchmark


def autopandas_01():
  # Turned the desired indices [0, 2, 4] into an input.
  examples = [
      benchmark.Example(
          inputs=[
              [[5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
              [0, 2, 4],
          ],
          output=[[5, 7, 9], [10, 12, 14]]
      ),
  ]
  constants = []
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.gather(in1, in2, axis=1)'
  source = 'SO_11881165_depth1'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_01')


def autopandas_02():
  """AutoPandas benchmark.

  The DataFrame input:
                 value1  value2
  group1 group2
  a      c          1.1     7.1
         c          2.0     8.0
         d          3.0     9.0
  b      d          4.0    10.0
         d          5.0    11.0
         e          6.0    12.0

  The DataFrame output:
          value1  value2
  group2
  c          1.1     7.1
  c          2.0     8.0
  d          3.0     9.0

  Notice that "c", "d", and "e" are just treated as data, so we'll replace them
  with 1, 2, and 3, respectively. "group1" acts as a third dimension, so our
  input tensor will be 3D. We'll also make "group1" the innermost axis, to make
  the problem not super trivial in TensorFlow.

  Finally, I changed some numbers to rule out tf.reduce_min(axis=2).
  """
  examples = [
      benchmark.Example(
          inputs=[
              [[[1, 2], [1, 2], [2, 3]],  # group2
               [[1.1, 4], [2, 1], [3, 6]],  # value1
               [[7.1, 10], [8, 11], [9, 7]]],  # value2
          ],
          output=[[1, 1, 2],
                  [1.1, 2, 3],
                  [7.1, 8, 9]]
      ),
  ]
  constants = []
  description = ''  # No description for AutoPandas benchmarks.
  # Note that we don't have ops for indexing/slicing axis 2. But, TF-Coder will
  # find a workaround.
  target_program = 'in1[:, :, 0]'
  source = 'SO_11941492_depth1'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_02')


def autopandas_03():
  # The "series" and "step" columns only serve to identify where the data should
  # move to. The number of unique "series" and "step" values determine the
  # output's shape, so we provide them (3 and 5) as constants. The real data is
  # in the "value" column, so it's a 1D tensor. This is simply a tensor reshape
  # and transpose.
  examples = [
      benchmark.Example(
          inputs=[
              [1000, 1001, 1002, 1003, 1004, 1005, 1006, 1007, 1008, 1009, 1010,
               1011, 1012, 1013, 1014],
          ],
          output=[[1000, 1003, 1006, 1009, 1012],
                  [1001, 1004, 1007, 1010, 1013],
                  [1002, 1005, 1008, 1011, 1014]]
      ),
  ]
  constants = [3, 5]
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.transpose(tf.reshape(in1, (5, 3)))'
  source = 'SO_13647222_depth1'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_03')


def autopandas_04():
  # This problem boils down to "select rows of the table where row['line_race']
  # is nonzero". We use a simpler example to describe the same problem idea.
  examples = [
      benchmark.Example(
          inputs=[
              [[1, 2, 3, 4, 5],
               [9, 8, 7, 6, 5],
               [3, 0, 2, 5, 8],
               [8, 8, 6, 3, 2],
               [2, 0, 7, 7, 3],
               [9, 0, 3, 2, 7],
               [1, 3, 8, 9, 4]],
          ],
          output=[[1, 2, 3, 4, 5],
                  [9, 8, 7, 6, 5],
                  [8, 8, 6, 3, 2],
                  [1, 3, 8, 9, 4]],
      ),
  ]
  constants = []
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.boolean_mask(in1, tf.cast(in1[:, 1], tf.bool))'
  source = 'SO_18172851_depth1'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_04')


def autopandas_05():
  # The problem is to sort a table using values in one column. Their code that
  # constructs the output actually sorts by 3 columns, but their example has
  # unique values for the first column, so the other two columns don't affect
  # the result at all. Instead of having dates and strings and numbers, we just
  # use numbers. Sort by column 0 in increasing order.
  examples = [
      benchmark.Example(
          inputs=[
              [[6, 3, 7, 8, 4],
               [8, 9, 4, 5, 3],
               [1, 5, 3, 6, 9],
               [2, 1, 4, 3, 2],
               [7, 9, 6, 2, 7],
               [5, 8, 0, 4, 2]],
          ],
          output=[[1, 5, 3, 6, 9],
                  [2, 1, 4, 3, 2],
                  [5, 8, 0, 4, 2],
                  [6, 3, 7, 8, 4],
                  [7, 9, 6, 2, 7],
                  [8, 9, 4, 5, 3]]
      ),
  ]
  constants = []
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.gather(in1, tf.argsort(in1[:, 0], stable=True))'
  source = 'SO_49583055_depth1'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_05')


# SO_49592930_depth1 is out of scope. It involves taking the union of two
# mappings from keys to values, where one such mapping takes precedence over the
# other in case of overlapping keys. TensorFlow is not designed to handle
# mappings, and I (Kensen) cannot think of a way to do this using TensorFlow.

# SO_49572546_depth1 is exactly the same as the one above, except for the
# ordering of two inputs and the data. It's out of scope for the same reason.


@benchmark.ignore("This isn't in AutoPandas's table of results, idk why")
def autopandas_06_ignored():
  """AutoPandas benchmark.

  The input DataFrame:
      X   Y   Z
  4  X1  Y2  Z3
  5  X1  Y1  Z1
  6  X1  Y1  Z1
  7  X1  Y1  Z2

  The output DataFrame:
  Z    Z1   Z2   Z3
  Y
  Y1  1.0  1.0  NaN
  Y2  NaN  NaN  1.0

  Basically, the Ys are row indices and the Zs are column indices. The X1 does
  not really matter, we just want a boolean output table (1.0=True, NaN=False)
  that identifies which Y and Z pairs appeared in the input table.

  Their example is tiny so I expanded it.

  This task also doesn't appear in their table of results?
  """
  examples = [
      benchmark.Example(
          inputs=[
              # Y  Z
              [[1, 2],
               [0, 0],
               [0, 0],
               [0, 1],
               [4, 2],
               [4, 3],
               [4, 2],
               [2, 1]],
          ],
          output=[[True, True, False, False],
                  [False, False, True, False],
                  [False, True, False, False],
                  [False, False, False, False],
                  [False, False, True, True]]
      ),
  ]
  constants = []
  description = ''  # No description for AutoPandas benchmarks.
  # TODO(kshi): Note that tf.scatter_nd is not yet supported by TF-Coder!!
  target_program = 'tf.cast(tf.scatter_nd(in1, updates=tf.ones(8), shape=[5, 4]), tf.bool)'
  source = 'SO_12860421_depth1'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_06_ignored')


def autopandas_06():
  # This is very similar to autopandas_03, they're both pivot tables.
  examples = [
      benchmark.Example(
          inputs=[
              [4, 5, 6, 7],
          ],
          output=[[4, 5], [6, 7]]
      ),
  ]
  constants = [2]
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.reshape(in1, (2, 2))'
  source = 'SO_13261175_depth1'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_06')


# SO_13793321_depth1 is out of scope. It's merging two tables, keeping the rows
# with matching values in a particular column. This isn't something TensorFlow
# was designed for.


def autopandas_07():
  # This is exactly the same as autopandas_05 (sort rows based on a column), but
  # with different data. Like in the other task, their code for creating the
  # output actually sorts using multiple columns, but their data contains unique
  # values. I'm just going to mash my keyboard again to get different data.
  examples = [
      benchmark.Example(
          inputs=[
              [[8, 5, 9, 3],
               [3, 6, 6, 8],
               [1, 2, 3, 4],
               [7, 6, 5, 4],
               [4, 7, 2, 6]]
          ],
          output=[[1, 2, 3, 4],
                  [3, 6, 6, 8],
                  [4, 7, 2, 6],
                  [7, 6, 5, 4],
                  [8, 5, 9, 3]]
      ),
  ]
  constants = []
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.gather(in1, tf.argsort(in1[:, 0], stable=True))'
  source = 'SO_14085517_depth1'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_07')


def autopandas_08():
  # I added 3 extra rows so that the solution isn't just slicing, and chose the
  # numbers to test the boundary of the condition row[0] > 1. The boundary (1)
  # is given as a constant.
  examples = [
      benchmark.Example(
          inputs=[
              [[5, 7], [6, 8], [-1, 9], [-2, 10], [2, 11], [1, 12], [3, -3]],
          ],
          output=[[5, 7], [6, 8], [2, 11], [3, -3]]
      ),
  ]
  constants = [1]
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.boolean_mask(in1, tf.greater(in1[:, 0], 1))'
  source = 'SO_11418192_depth2'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_08')


# SO_49567723_depth2 is out of scope. It's another table merge.

# SO_49987108_depth2 is out of scope. It uses DataFrame.fillna(method='ffill'),
# which has no equivalent (afaik) in TensorFlow.


def autopandas_09():
  # This is yet another sort. But this time, the data actually does require
  # sorting by both columns! I have constructed the example to reflect this.
  # Also, this problem (sort by 2 columns) is the same as stackoverflow_19.
  examples = [
      benchmark.Example(
          inputs=[
              [[8, 5, 9, 3],
               [3, 6, 6, 8],
               [7, 9, 3, 4],
               [7, 6, 5, 4],
               [3, 7, 2, 6]]
          ],
          output=[[3, 6, 6, 8],
                  [3, 7, 2, 6],
                  [7, 6, 5, 4],
                  [7, 9, 3, 4],
                  [8, 5, 9, 3]]
      ),
  ]
  constants = []
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.gather(tf.gather(in1, tf.argsort(in1[:, 1], stable=True)), tf.argsort(tf.gather(in1, tf.argsort(in1[:, 1], stable=True))[:, 0], stable=True))'
  source = 'SO_13261691_depth2'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_09')


# SO_13659881_depth2 is out of scope. It involves counting how many times each
# row appears, and then appending that count to the deduplicated rows.
# TensorFlow is not designed for deduplicating rows.


def autopandas_10():
  # Drop NaN. NaN is given as a constant.
  examples = [
      benchmark.Example(
          inputs=[
              [float('nan'), 11, 12, float('nan'), 16, 18],
          ],
          output=[11, 12, 16, 18]
      ),
  ]
  constants = [float('nan')]
  description = ''  # No description for AutoPandas benchmarks.
  # TODO(kshi): We don't support tf.math.is_nan() and tf.math.logical_not()!!
  # Once again, TF-Coder is better than me. TF-Coder replaces
  # `tf.math.logical_not(tf.math.is_nan(in1))` with `tf.equal(in1, in1)` because
  # NaN is the only thing that's not equal to itself.
  target_program = 'tf.cast(tf.boolean_mask(in1, tf.math.logical_not(tf.math.is_nan(in1))), tf.int32)'
  source = 'SO_13807758_depth2'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_10')


# SO_34365578_depth2 is out of scope. It involves grouping values and doing a
# sum aggregation for each group. TensorFlow is not designed for grouping values
# like this.

# SO_10982266_depth3 is also out of scope, involving a grouping and then mean
# aggregation for each group.


def autopandas_11():
  # Transpose and prepend the row index. I changed the data to not have such
  # obvious patterns.
  examples = [
      benchmark.Example(
          inputs=[
              [[1, 4, 2, 7, 6], [20, 10, 50, 40, 30]],
          ],
          output=[[0, 1, 20],
                  [1, 4, 10],
                  [2, 2, 50],
                  [3, 7, 40],
                  [4, 6, 30]]
      ),
  ]
  constants = []
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.transpose(tf.concat((tf.expand_dims(tf.range(5), axis=0), in1), axis=0))'
  source = 'SO_11811392_depth3'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_11')


def autopandas_12():
  # For each pair, compute pair[1] / (pair[0] + pair[1]).
  # I added some numbers to avoid in1[:, :, 1] == tf.reduce_max(in1, axis=-1).
  examples = [
      benchmark.Example(
          inputs=[
              [[[2, 8], [2, 6], [6, 2]],
               [[0, 2], [1, 1], [2, 0]]],
          ],
          output=[[0.8, 0.75, 0.25],
                  [1.0, 0.5, 0.0]]
      ),
  ]
  constants = []
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.cast(tf.divide(in1[:, :, 1], tf.reduce_sum(in1, axis=2)), tf.float32)'
  source = 'SO_49581206_depth3'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_12')


def autopandas_13():
  # Select rows where row[1] is in some set. I'm matching their data where
  # possible.
  examples = [
      benchmark.Example(
          inputs=[
              [[101, 0, 11, 0],
               [102, 1, 12, 4],
               [103, 2, 13, 2],
               [104, 3, 14, 8],
               [105, 4, 15, 4],
               [106, 5, 16, 5],
               [107, 6, 17, 4],
               [108, 7, 18, 7],
               [109, 8, 19, 7],
               [110, 9, 20, 4]],
              [4, 2, 6],
          ],
          output=[[103, 2, 13, 2],
                  [105, 4, 15, 4],
                  [107, 6, 17, 4]]
      ),
  ]
  constants = []
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.boolean_mask(in1, tf.reduce_any(tf.equal(in1[:, 1], tf.expand_dims(in2, axis=1)), axis=0))'
  source = 'SO_12065885_depth3'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_13')


def autopandas_14():
  # The intended solution is to do a merge, but the example actually has nothing
  # to merge, so all that happens is an extra column is created filled with NaN.
  # Why are there so many problems where the example completely underspecifies
  # the intended transformation? Instead of ignoring the problem because it has
  # a merge (out of scope), let's just try to append a NaN column.
  examples = [
      benchmark.Example(
          inputs=[
              [[1, 0, 1, 2],
               [1, 1, 3, 4],
               [2, 0, 1, 2],
               [2, 1, 3, 4]],
              # They have another input that is used to provide the column
              # header for the new NaN-filled column. Tensors don't have column
              # headers so this input is useless. Omit it, or else TF-Coder will
              # be required to use it.
          ],
          output=[[1, 0, 1, 2, float('nan')],
                  [1, 1, 3, 4, float('nan')],
                  [2, 0, 1, 2, float('nan')],
                  [2, 1, 3, 4, float('nan')]]
      ),
  ]
  constants = [float('nan')]
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.concat((tf.cast(in1, tf.float32), tf.fill([4, 1], float("nan"))), axis=1)'
  source = 'SO_13576164_depth3'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_14')


# SO_14023037_depth3 is out of scope. It uses DataFrame.fillna(method='bfill'),
# which has no equivalent (afaik) in TensorFlow.


def autopandas_15():
  # The task is to group by everything except one numeric column, and then
  # cumsum over that numeric column. Grouping is out-of-scope, and the groups
  # essentially become row headers anyway. The only part of this that is doable
  # in TensorFlow is the cumsum part. I changed the example because it was too
  # simple.
  examples = [
      benchmark.Example(
          inputs=[
              [1, 1, 2, 1, 3, 2],
          ],
          output=[1, 2, 4, 5, 8, 10]
      ),
  ]
  constants = []
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.cumsum(in1)'
  source = 'SO_53762029_depth3'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_15')


# SO_21982987_depth3 is out of scope, involving a grouping and then mean
# aggregation for each group.


def autopandas_16():
  # Reduce mean across columns, and then ignore column 0.
  examples = [
      benchmark.Example(
          inputs=[
              [[0, 6, 0],
               [3, 101, 14],
               [0, 91, 6],
               [5, 15, 0]],
          ],
          output=[53.25, 5.00]
      ),
  ]
  constants = []
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'tf.reduce_mean(tf.cast(in1, tf.float32), axis=0)[1:]'
  source = 'SO_39656670_depth3'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_16')


# SO_23321300_depth3 is out of scope, involving a grouping and then mean
# aggregation for each group.


# A template for easy copy/pasting. Copying an existing benchmark and replacing
# parts of it will lead to a state where the benchmark is half-correct, but not
# obviously so. Copy this template instead when creating new benchmarks.
"""

def autopandas_NUMBER():
  examples = [
      benchmark.Example(
          inputs=[
              INPUT_1,
              INPUT_2,
          ],
          output=OUTPUT
      ),
  ]
  constants = [CONSTANTS]
  description = ''  # No description for AutoPandas benchmarks.
  target_program = 'SOLUTION_PROGRAM'
  source = 'PROBLEM_SOURCE'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='autopandas_NUMBER')

"""  # pylint: disable=pointless-string-statement



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for operation_statistics.py."""

from absl.testing import absltest
from tf_coder.value_search import operation_statistics


class OperationStatisticsTest(absltest.TestCase):

  def test_update(self):
    statistics = operation_statistics.OperationStatistics()
    statistics.update('a', count=10, successes=1, time=1.5)
    statistics.update('b', count=80, successes=60, time=40.0)
    statistics.update('a', count=100, successes=10, time=8.5)
    statistics.update('c', count=0, successes=0, time=0.0)
    self.assertEqual(statistics.total_apply_count, 190)
    self.assertEqual(statistics.total_apply_successes, 71)
    self.assertEqual(statistics.operation_apply_time,
                     {'a': 10.0, 'b': 40.0, 'c': 0.0})
    self.assertEqual(statistics.operation_apply_count,
                     {'a': 110, 'b': 80, 'c': 0})
    self.assertEqual(statistics.operation_apply_successes,
                     {'a': 11, 'b': 60, 'c': 0})
    self.assertEqual(statistics.all_operation_names, {'a', 'b', 'c'})

  def test_get_total_time(self):
    statistics = operation_statistics.OperationStatistics()
    statistics.update('a', count=10, successes=1, time=1.5)
    statistics.update('b', count=80, successes=60, time=40.0)
    statistics.update('a', count=100, successes=10, time=8.5)
    statistics.update('c', count=0, successes=0, time=0.0)
    self.assertEqual(statistics.get_total_time(), 50.0)

  def test_statistics_as_string(self):
    statistics = operation_statistics.OperationStatistics()
    statistics.update('a', count=10, successes=1, time=1.5)
    statistics.update('b', count=80, successes=60, time=40.0)
    statistics.update('a', count=100, successes=10, time=8.5)
    statistics.update('c', count=0, successes=0, time=0.0)
    statistics.update('do_not_output', count=100, successes=10, time=1.23)

    result = statistics.statistics_as_string(operation_names=['a', 'b', 'c'],
                                             num_unique_values=77,
                                             elapsed_time=100.0)

    a_row = operation_statistics._ROW_FORMAT_STR.format(
        name='a', eps=11.0, sps=1.1, executions=110, successes=11, rate=0.1,
        time=10.0, time_frac=0.2)
    self.assertIn(a_row, result)
    b_row = operation_statistics._ROW_FORMAT_STR.format(
        name='b', eps=2.0, sps=1.5, executions=80, successes=60, rate=0.75,
        time=40.0, time_frac=0.8)
    self.assertIn(b_row, result)
    nan = float('NaN')
    c_row = operation_statistics._ROW_FORMAT_STR.format(
        name='c', eps=nan, sps=nan, executions=0, successes=0, rate=nan,
        time=0.0, time_frac=0.0)
    self.assertIn(c_row, result)
    self.assertNotIn('do_not_output', result)

    self.assertIn('Number of evaluations: 290', result)
    self.assertIn('Number of successful evaluations: 81', result)
    self.assertIn('Total time applying operations: 51.23 sec', result)
    self.assertIn('Number of unique values: 77', result)
    self.assertIn('Executions per second: 2.9', result)

  def test_statistics_as_string_sorted_by_time(self):
    statistics = operation_statistics.OperationStatistics()
    statistics.update('fast_op', count=10, successes=1, time=1.5)
    statistics.update('slow_op', count=10, successes=1, time=2.5)
    result = statistics.statistics_as_string(sort_by_time=False)
    self.assertLess(result.index('fast_op'), result.index('slow_op'))
    result = statistics.statistics_as_string(sort_by_time=True)
    self.assertLess(result.index('slow_op'), result.index('fast_op'))

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for filter_group.py."""

from absl.testing import absltest
from tf_coder import filter_group


class FilterGroupTest(absltest.TestCase):

  def test_filter_group_enum_names(self):
    for enum_value in filter_group.FilterGroup:
      self.assertEqual(enum_value.name, enum_value.value)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A Benchmark class for TF-Coder."""

import collections


# A list of input tensors along with the expected output tensor.
Example = collections.namedtuple('Example', ['inputs', 'output'])


# The number of unnamed Benchmark objects created so far (a global variable).
#
# This is used to assign distinct names to unnamed benchmarks. However, the
# assigned name can clash with another explicitly-provided name. Unique names
# are preferable for easier testing and debugging. However, uniqueness is not
# required or enforceable, since any name can be provided to the constructor.
_num_unnamed_benchmarks = 0


class Benchmark(object):
  """A class for specifying the benchmarks (example problems) for TF-Coder.

  Attributes:
    examples: A nonempty list of Example namedtuples.
    constants: A list of constants.
    description: A natural-language description of the task.
    target_program: A string representing the desired program, if known.
    source: A string describing the source of the problem, preferably a
      hyperlink if available.
    name: A name (string) for the benchmark.
    num_inputs: The number of inputs in each example.
    should_ignore: A boolean indicating whether the benchmark should be ignored,
      perhaps because it is unsuitable or unreasonable for synthesis.
    ignore_reason: A string describing the reason for ignoring the benchmark, if
      should_ignore is True.
  """

  def __init__(self, examples, constants=None, description=None,
               target_program=None, source=None, name=None):
    """Initializes a Benchmark.

    Args:
      examples: A nonempty list of user-provided Example namedtuples.
      constants: A list of user-provided constants.
      description: A natural-language description of the task.
      target_program: A string representing the desired program, if known.
      source: A string describing the source of the problem, preferably a
        hyperlink if available.
      name: A name (string) for the benchmark. If not provided, a name will be
        assigned.

    Raises:
      ValueError: If there are no examples, or an example has no output, or the
      examples have inconsistent numbers of inputs.
    """
    if not examples:
      raise ValueError('A benchmark must have at least 1 example.')

    self.examples = examples
    self.constants = [] if constants is None else constants
    self.description = '' if description is None else description
    self.target_program = target_program
    self.source = source

    if name is None:
      global _num_unnamed_benchmarks
      _num_unnamed_benchmarks += 1
      self.name = 'unnamed_benchmark_' + str(_num_unnamed_benchmarks)
    else:
      self.name = name

    # Examples must have the same number of inputs and non-None output.
    self.num_inputs = len(examples[0].inputs)
    for example in examples:
      if example.output is None:
        raise ValueError('Examples must have non-None output.')
      if len(example.inputs) != self.num_inputs:
        raise ValueError('The examples in a bechmark must have the same number '
                         'of inputs.')

    # Set by the @ignore decorator.
    self.should_ignore = False
    self.ignore_reason = None


def ignore(ignore_reason):
  """A decorator to be applied to benchmark-creating functions.

  The benchmark returned by the decorated function will have should_ignore set
  to True and ignore_reason set as the decorator's argument.

  Args:
    ignore_reason: A string describing the reason for ignoring the benchmark.

  Returns:
    A decorator that adds the supplied ignore_reason to the benchmark.
  """
  def add_ignore_reason(benchmark_function):
    def wrapper():
      benchmark = benchmark_function()
      benchmark.should_ignore = True
      benchmark.ignore_reason = ignore_reason
      return benchmark
    return wrapper
  return add_ignore_reason



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A model that predicts operations from features of the I/O tensors."""

import collections
import json
import os
import timeit
from typing import Any, Dict, Text, Tuple

from absl import logging
import six
import tensorflow as tf
import termcolor
from tf_coder.datasets import collect_tensor_data


ConfigType = Dict[Text, Any]

Model = collections.namedtuple('Model', [
    'optimizer',

    'kind_embeddings',  # num_kinds, embedding_size
    'dtype_embeddings',  # num_dtypes, embedding_size
    'rank_embeddings',  # max_rank + 1, embedding_size
    'shape_buckets_embeddings',  # num_shape_buckets, embedding_size
    'float_buckets_embeddings',  # num_float_buckets, embedding_size
    'count_buckets_embeddings',  # num_count_buckets, embedding_size
    'booleans_embeddings',  # 2, embedding_size
    'io_comparisons_embeddings',  # 3, embedding_size
    'io_booleans_embeddings',  # 2, embedding_size
    'io_count_buckets_embeddings',  # num_count_buckets, embedding_size
    'num_inputs_embeddings',  # max_num_inputs, embedding_size

    'feedforward_layers',  # each (batch_size, example_len or feedforward_size
                           #       -> batch_size, feedforward_size)
    'output_layer',  # batch_size, feedforward_size -> batch_size, num_ops
])
Result = collections.namedtuple('Result', [
    'operation_logits',  # batch_size, num_ops
])
Example = collections.namedtuple('Example',  # pylint: disable=invalid-name
                                 list(collect_tensor_data.FEATURE_NAME_TO_TYPE))


def parse_example_proto(serialized_example: Text) -> Dict[Text, tf.Tensor]:
  """Parses the serialized tf.train.Example proto."""
  features = {}
  for feature_name, feature_type in six.iteritems(
      collect_tensor_data.FEATURE_NAME_TO_TYPE):
    dtype = (
        tf.int64 if feature_type == collect_tensor_data.FeatureType.INT else  # pylint: disable=g-long-ternary
        tf.float32 if feature_type == collect_tensor_data.FeatureType.FLOAT else
        tf.string if feature_type == collect_tensor_data.FeatureType.STRING else
        None)
    assert dtype is not None
    features[feature_name] = tf.io.VarLenFeature(dtype)
  parsed = tf.io.parse_single_example(serialized_example, features)
  for key in parsed:
    parsed[key] = tf.sparse.to_dense(parsed[key])
  return parsed


def get_model(config: ConfigType) -> Model:
  """Gets a namedtuple with all the variables and layers of the model."""
  initializer = tf.initializers.GlorotUniform()
  optimizer = tf.keras.optimizers.Adam(config['learning_rate'])
  optimizer.iterations = tf.compat.v1.train.get_or_create_global_step()

  return Model(
      optimizer=optimizer,

      kind_embeddings=tf.Variable(
          name='kind_embeddings',
          initial_value=initializer(
              shape=[config['num_kinds']] * 2,
              dtype=tf.float32)),
      dtype_embeddings=tf.Variable(
          name='dtype_embeddings',
          initial_value=initializer(
              shape=[config['num_dtypes']] * 2,
              dtype=tf.float32)),
      rank_embeddings=tf.Variable(
          name='rank_embeddings',
          initial_value=initializer(
              shape=[config['max_rank'] + 1] * 2,
              dtype=tf.float32)),
      shape_buckets_embeddings=tf.Variable(
          name='shape_buckets_embeddings',
          initial_value=initializer(
              shape=[config['num_shape_buckets']] * 2,
              dtype=tf.float32)),
      float_buckets_embeddings=tf.Variable(
          name='float_buckets_embeddings',
          initial_value=initializer(
              shape=[config['num_float_buckets']] * 2,
              dtype=tf.float32)),
      count_buckets_embeddings=tf.Variable(
          name='count_buckets_embeddings',
          initial_value=initializer(
              shape=[config['num_count_buckets']] * 2,
              dtype=tf.float32)),
      booleans_embeddings=tf.Variable(
          name='booleans_embeddings',
          initial_value=initializer(
              shape=[2] * 2,
              dtype=tf.float32)),
      io_comparisons_embeddings=tf.Variable(
          name='io_comparisons_embeddings',
          initial_value=initializer(
              shape=[3] * 2,
              dtype=tf.float32)),
      io_booleans_embeddings=tf.Variable(
          name='io_booleans_embeddings',
          initial_value=initializer(
              shape=[2] * 2,
              dtype=tf.float32)),
      io_count_buckets_embeddings=tf.Variable(
          name='io_count_buckets_embeddings',
          initial_value=initializer(
              shape=[config['num_count_buckets']] * 2,
              dtype=tf.float32)),
      num_inputs_embeddings=tf.Variable(
          name='num_inputs_embeddings',
          initial_value=initializer(
              shape=[config['max_num_inputs'] + 1] * 2,
              dtype=tf.float32)),

      feedforward_layers=[tf.keras.layers.Dense(config['feedforward_size'])
                          for _ in range(config['num_feedforward_layers'])],
      output_layer=tf.keras.layers.Dense(config['num_ops']),
  )


@tf.function
def predict(model: Model, example: Example) -> Result:
  """Predict which ops will be used.

  Args:
    model: The Model (a namedtuple containing the variables used by the model).
    example: The Example to do prediction for. An example represents an I/O
      tensor pair.

  Returns:
    A Result object with the prediction results.
  """
  embedding_pairs = [
      (model.kind_embeddings, example.kind),
      (model.dtype_embeddings, example.dtype),
      (model.rank_embeddings, example.rank),
      (model.shape_buckets_embeddings, example.shape_buckets),
      (model.float_buckets_embeddings, example.float_buckets),
      (model.count_buckets_embeddings, example.count_buckets),
      (model.booleans_embeddings, example.booleans),
      (model.io_comparisons_embeddings, example.io_comparisons),
      (model.io_booleans_embeddings, example.io_booleans),
      (model.io_count_buckets_embeddings, example.io_count_buckets),
      (model.num_inputs_embeddings, example.num_inputs),
  ]

  embeddings = [tf.nn.embedding_lookup(embeddings, item)
                for embeddings, item in embedding_pairs]

  unembedded = [
      example.shape,
      # example.floats,  # May include inf and nan! Use the buckets only.
      example.counts,
      example.fractions,
      example.io_counts,
      example.io_fractions,
  ]

  # Some features are single numbers, and some are sequences of numbers. Some
  # features are embedded, and some are left unembedded. In any case,
  # concatenate everything into one long input vector.
  batch_size = example.kind.shape[0]
  flattened_embeddings = [tf.reshape(embedding, (batch_size, -1))
                          for embedding in embeddings]
  to_concat = [tf.cast(tensor, tf.float32)
               for tensor in flattened_embeddings + unembedded]
  batched_inputs = tf.concat(to_concat, axis=1)

  layer_output = batched_inputs
  for feedforward_layer in model.feedforward_layers:
    layer_output = feedforward_layer(layer_output)
  operation_logits = model.output_layer(layer_output)

  return Result(operation_logits=operation_logits)


@tf.function
def f_beta_loss(ground_truth, predictions, beta):
  """Computes F_beta loss (1 minus F_beta score) in a differentiable way."""
  # Inspired by:
  # https://www.kaggle.com/rejpalcz/best-loss-function-for-f1-score-metric
  true_positives = tf.reduce_sum(ground_truth * predictions)
  false_positives = tf.reduce_sum(
      tf.where(tf.equal(ground_truth, 0.0), 1.0, 0.0) * predictions)
  false_negatives = tf.reduce_sum(ground_truth * (1 - predictions))

  epsilon = 0.0001
  precision = true_positives / (true_positives + false_positives + epsilon)
  recall = true_positives / (true_positives + false_negatives + epsilon)

  beta_squared = beta ** 2
  f_beta_score = ((1 + beta_squared) * precision * recall /
                  (beta_squared * precision + recall + epsilon))
  return 1 - f_beta_score


@tf.function
def get_loss(model: Model,
             config: ConfigType,
             result: Result,
             example: Example) -> float:
  """Computes the loss for the model."""
  del model  # Unused.

  # The ground truth.
  operations = tf.clip_by_value(tf.cast(example.operations, tf.float32), 0, 1)

  if config['weighted_ops']:
    op_counts = tf.cast(config['op_counts'], tf.float32)
    if config['weight_numerator'] == 'mean':
      weight_numerator = tf.reduce_mean(op_counts)
    elif config['weight_numerator'] == 'max':
      weight_numerator = tf.reduce_max(op_counts)
    else:
      raise ValueError("Unknown config['weight_numerator']: {}".format(
          config['weight_numerator']))
    weight_if_op_used = tf.minimum(
        tf.cast(config['max_weight'], tf.float32),
        weight_numerator / op_counts)
    weights = tf.where(tf.cast(operations, tf.bool),  # [batch_size, num_ops]
                       weight_if_op_used,  # [num_ops]
                       1.0)

  loss_type = config['loss']
  if loss_type == 'sigmoid_ce':
    loss_per_op = tf.nn.sigmoid_cross_entropy_with_logits(
        labels=operations, logits=result.operation_logits)
    if config['weighted_ops']:
      loss_per_op *= weights
    return tf.reduce_mean(loss_per_op)
  elif loss_type == 'f_beta':
    if config['weighted_ops']:
      # `operations` now contains 0 if the op is not used, otherwise a real
      # number representing the number of times we should treat the op as used.
      operations *= weights
    return f_beta_loss(ground_truth=operations,
                       predictions=tf.sigmoid(result.operation_logits),
                       beta=config['beta'])
  else:
    raise ValueError("Unhandled config['loss']: {}".format(loss_type))


@tf.function
def get_accuracy(model: Model,
                 result: Result,
                 example: Example) -> Tuple[float, float, float, float, float]:
  """Returns a tuple containing various accuracy metrics."""
  del model  # Unused.

  # TODO(kshi): Refactor to reuse the simpler code in f_beta_loss.

  predictions = tf.cast(tf.greater(tf.sigmoid(result.operation_logits), 0.5),
                        tf.int32)
  operations = tf.clip_by_value(tf.cast(example.operations, tf.int32), 0, 1)

  correct = tf.equal(predictions, operations)
  num_correct = tf.reduce_sum(tf.cast(correct, tf.int32))
  accuracy = num_correct / tf.size(correct)

  # Precision.
  # Among ops that the model predicted were used, how often was it correct?
  predicted_positive_indices = tf.where(tf.equal(predictions, 1))
  predicted_positive_truth = tf.gather_nd(operations,
                                          predicted_positive_indices)
  num_correct = tf.reduce_sum(predicted_positive_truth)
  precision = num_correct / tf.size(predicted_positive_truth)

  # Recall, i.e., true positive rate.
  # Among ops that were actually used, how often was the model correct?
  actually_positive_indices = tf.where(tf.equal(operations, 1))
  actually_positive_predictions = tf.gather_nd(predictions,
                                               actually_positive_indices)
  num_correct = tf.reduce_sum(actually_positive_predictions)
  recall = num_correct / tf.size(actually_positive_predictions)

  # True negative rate.
  # Among ops that were actually not used, how often was the model correct?
  actually_negative_indices = tf.where(tf.equal(operations, 0))
  actually_negative_predictions = tf.gather_nd(predictions,
                                               actually_negative_indices)
  num_correct = tf.reduce_sum(1 - actually_negative_predictions)
  true_negative_rate = num_correct / tf.size(actually_negative_predictions)

  # F1 score.
  f1_score = 2 * precision * recall / (precision + recall)

  return accuracy, precision, recall, true_negative_rate, f1_score


def get_trainable_variables(model: Model):
  return [
      model.kind_embeddings,
      model.dtype_embeddings,
      model.rank_embeddings,
      model.shape_buckets_embeddings,
      model.float_buckets_embeddings,
      model.count_buckets_embeddings,
      model.booleans_embeddings,
      model.io_comparisons_embeddings,
      model.io_booleans_embeddings,
      model.io_count_buckets_embeddings,
      model.num_inputs_embeddings,
  ] + (sum([layer.variables for layer in model.feedforward_layers], []) +
       model.output_layer.variables)


def save_config(config, filepath):
  json_string = json.dumps(config, sort_keys=True, indent=4,
                           separators=(',', ': '))
  with tf.io.gfile.GFile(filepath, mode='w') as f:
    f.write(json_string)


def load_config(filepath):
  with tf.io.gfile.GFile(filepath, mode='r') as f:
    return json.load(f)


def create_checkpoint(model, global_step=None):
  """Creates a Checkpoint object."""
  trackables = model._asdict()
  if global_step is not None:
    trackables['global_step'] = global_step
  # Unpack the layers in the list of feedforward layers, since a list of layers
  # is not trackable.
  for i, layer in enumerate(trackables['feedforward_layers']):
    trackables['feedforward_layer_{}'.format(i + 1)] = layer
  del trackables['feedforward_layers']
  return tf.train.Checkpoint(**trackables)


def train(model: Model, train_dataset, eval_dataset, config, work_unit_dir):
  """Trains a model that predicts operations from I/O tensor features."""
  global_step = model.optimizer.iterations
  checkpoint_dir = os.path.join(work_unit_dir, 'checkpoints')

  config_filepath = os.path.join(work_unit_dir, 'config.json')
  save_config(config, config_filepath)

  log_dir = os.path.join(work_unit_dir, 'train_logs')
  writer = tf.summary.create_file_writer(log_dir, flush_millis=10000)

  checkpoint_path = tf.train.latest_checkpoint(checkpoint_dir)
  checkpoint = create_checkpoint(model, global_step)
  checkpoint_manager = tf.train.CheckpointManager(
      checkpoint, directory=checkpoint_dir, max_to_keep=1,
      keep_checkpoint_every_n_hours=config['keep_checkpoint_every_n_hours'])

  skip_epoch_checkpoint = False
  if checkpoint_path is not None:
    if config['allow_restore']:
      logging.info(termcolor.colored('Restoring from checkpoint: %s',
                                     color='red', attrs=['bold']),
                   checkpoint_path)
      checkpoint.restore(checkpoint_path)
      skip_epoch_checkpoint = True
    else:
      raise RuntimeError(
          'A preexisting checkpoint was found in the run directory. Aborting.')

  # Get the first batch from the eval dataset to use as the eval batch.
  eval_batch = Example(**next(iter(eval_dataset)))

  trainable_variables = None
  last_batch_end_time = timeit.default_timer()
  with writer.as_default():
    for epoch in range(1, 1 + config['num_epochs']):
      logging.info(
          termcolor.colored('Epoch %d', color='green', attrs=['bold']),
          epoch)

      if skip_epoch_checkpoint:
        skip_epoch_checkpoint = False
      else:
        checkpoint_manager.save()

      for batch_index, example_dict in enumerate(train_dataset):
        example = Example(**example_dict)
        if epoch == 1 and batch_index == 0:
          logging.info(example)
        with tf.GradientTape() as tape:
          result = predict(model, example)
          loss = get_loss(model, config, result, example)
        accuracy, precision, recall, true_negative_rate, f1_score = (
            get_accuracy(model, result, example))

        if trainable_variables is None:
          # The variables only become available after the first forward pass.
          trainable_variables = get_trainable_variables(model)

        grads = tape.gradient(loss, trainable_variables)
        grads, global_norm = tf.clip_by_global_norm(grads,
                                                    config['global_norm_clip'])
        model.optimizer.apply_gradients(zip(grads, trainable_variables))

        if tf.equal(global_step % config['save_step_frequency'], 0):
          checkpoint_manager.save()

        evaluate = tf.equal(global_step % config['eval_step_frequency'], 0)
        if evaluate:
          eval_result = predict(model, eval_batch)
          eval_loss = get_loss(model, config, eval_result, eval_batch)
        else:
          eval_loss = None

        write_summaries = tf.equal(
            global_step % config['summary_step_frequency'], 0)
        if write_summaries:
          tf.summary.experimental.set_step(global_step)
          tf.summary.scalar('loss', loss)
          tf.summary.scalar('global_norm', global_norm)
          tf.summary.scalar('accuracy', accuracy)
          tf.summary.scalar('precision', precision)
          tf.summary.scalar('recall', recall)
          tf.summary.scalar('true_negative_rate', true_negative_rate)
          tf.summary.scalar('f1_score', f1_score)
          if eval_loss is not None:
            tf.summary.scalar('eval_loss', eval_loss)
          # for variable in trainable_variables:
          #   tf.summary.histogram(variable.name.rstrip(':0'), variable)

        logging.info(termcolor.colored('Epoch %s, batch %s, global_step %s',
                                       color='green', attrs=['bold']),
                     epoch, batch_index + 1, int(global_step))
        logging.info(termcolor.colored('Loss: %s', color='magenta'),
                     loss.numpy())
        if eval_loss is not None:
          logging.info(termcolor.colored('Eval loss: %s', color='magenta',
                                         attrs=['bold']),
                       eval_loss.numpy())
        logging.info(termcolor.colored('Global norm: %s', color='magenta'),
                     global_norm.numpy())
        logging.info(termcolor.colored('Accuracy: %.3f%%', color='blue'),
                     accuracy.numpy() * 100)
        logging.info(termcolor.colored('Precision: %.3f%%', color='cyan'),
                     precision.numpy() * 100)
        logging.info(termcolor.colored('Recall: %.3f%%', color='cyan'),
                     recall.numpy() * 100)
        logging.info(termcolor.colored('True negative rate: %.3f%%',
                                       color='blue'),
                     true_negative_rate.numpy() * 100)
        logging.info(termcolor.colored('F1 score: %.3f%%',
                                       color='blue'),
                     f1_score.numpy() * 100)
        batch_end_time = timeit.default_timer()
        logging.info(termcolor.colored('Batch time: %.2f sec',
                                       color='blue'),
                     batch_end_time - last_batch_end_time)
        last_batch_end_time = batch_end_time


def eval_single_example(model: Model, serialized_example: Text) -> Result:
  """Runs the model on one example."""
  example_dict = parse_example_proto(serialized_example)
  for key in example_dict:
    example_dict[key] = tf.expand_dims(example_dict[key], axis=0)
  example = Example(**example_dict)
  return predict(model, example)



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for value_search_utils.py."""

from absl import logging
from absl.testing import absltest
from absl.testing import parameterized
import tensorflow as tf
from tf_coder import tensor_limits as limits
from tf_coder.value_search import value
from tf_coder.value_search import value_search_utils


class ValueSearchUtilsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('acceptable_size', (1, 2, 3), True),
      ('too_many_elements', (10, 20, 30), False),
      ('one_dimension_too_long', (1, 1, 8000), False),
      ('too_many_dimensions', (1, 1, 1, 2, 3), False))
  def test_check_tensor_size(self, tensor_shape, expected_result):
    # check_tensor_size() expects a Value object, but is also called from
    # Value.__init__. We first create a Value object with an ok tensor size, and
    # then replace its wrapped value with the new tensor, so that we can check
    # the result of check_tensor_size() without it causing the Value constructor
    # to throw an exception.
    tensor = tf.zeros(tensor_shape)
    tensor_value = value.InputValue(tf.constant([1]), 'test')
    tensor_value.value = tensor
    tensor_value.cached_info = {}
    self.assertEqual(value_search_utils.check_tensor_size(tensor_value),
                     expected_result)

  def test_check_tensor_size_scalar_tensor(self):
    self.assertTrue(value_search_utils.check_tensor_size(
        value.ConstantValue(tf.constant(1))))

  @parameterized.named_parameters(
      ('acceptable_size',
       tf.SparseTensor(indices=[[0, 0, 1], [1, 2, 3]],
                       values=[8, 9],
                       dense_shape=[100, 200, 1000]),
       True),
      ('too_many_elements',
       tf.SparseTensor(
           indices=[[i] for i in range(limits.MAX_TENSOR_ELEMENTS + 1)],
           values=[1] * (limits.MAX_TENSOR_ELEMENTS + 1),
           dense_shape=[10000]),
       False),
      ('too_many_dimensions',
       tf.SparseTensor(indices=[[0] * (limits.MAX_NUM_DIMENSIONS + 1)],
                       values=[1],
                       dense_shape=[100] * (limits.MAX_NUM_DIMENSIONS + 1)),
       False))
  def test_check_sparse_tensor_size(self, sparse_tensor, expected_result):
    # check_tensor_size() expects a Value object, but is also called from
    # Value.__init__. We first create a Value object with an ok tensor size, and
    # then replace its wrapped value with the new tensor, so that we can check
    # the result of check_tensor_size() without it causing the Value constructor
    # to throw an exception.
    input_value = value.InputValue(
        tf.sparse.from_dense(tf.constant([[0, 1], [0, 0]])), 'test')
    input_value.value = sparse_tensor
    input_value.cached_info = {}
    self.assertEqual(value_search_utils.check_sparse_tensor_size(input_value),
                     expected_result)


if __name__ == '__main__':
  logging.set_verbosity(logging.ERROR)

  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for tfidf_handler.py."""

from absl.testing import absltest
from absl.testing import parameterized
import mock
import numpy as np
from tf_coder.benchmarks import benchmark as benchmark_module
from tf_coder.natural_language import tfidf_handler
from tf_coder.value_search import value_search_settings as settings_module


def _create_operation(name, docstring):
  """Creates a dummy operation with the given name and docstring."""
  # For the purposes of this test, an Operation is any object with `.name` and
  # `.metadata.docstring` attributes.
  operation = mock.Mock()
  operation.name = name
  operation.metadata = mock.Mock()
  operation.metadata.docstring = docstring
  return operation


class TfidfHandlerTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('tile', 'Tile a tensor multiple times', 'tf.tile(input, multiples)'),
      ('tensordot', 'Something like np.tensordot', 'tf.tensordot(a, b, axes)'))
  def test_score_description_best_operation(self, description, best_operation):
    handler = tfidf_handler.TfidfDescriptionHandler()
    score_dict = handler.score_description(description)
    self.assertTrue(all(0 <= value <= 1 for value in score_dict.values()))
    max_score = max(score_dict.values())
    self.assertEqual(score_dict[best_operation], max_score)

  def test_score_description_exact_values(self):
    # This example is taken from the following link:
    # https://scikit-learn.org/stable/modules/feature_extraction.html#tfidf-term-weighting
    counts = [[3, 0, 1],
              [2, 0, 0],
              [3, 0, 0],
              [4, 0, 0],
              [3, 2, 0],
              [3, 0, 2]]
    docstrings = [' '.join(['apple'] * i + ['banana'] * j + ['clementine'] * k)
                  for i, j, k in counts]
    operations = [_create_operation('operation_{}'.format(i), docstring)
                  for i, docstring in enumerate(docstrings)]
    handler = tfidf_handler.TfidfDescriptionHandler(operations=operations)

    # Same counts as docstrings[0]. "dragonfruit" is not in the vocabulary.
    description = 'apple clementine apple dragonfruit apple'

    scores = handler.score_description(description)
    sorted_scores = [scores[name] for name in sorted(scores.keys())]

    # Keep the formatting as in the link above. pylint: disable=bad-whitespace
    expected_term_document_matrix = [[0.85151335, 0.        , 0.52433293],
                                     [1.        , 0.        , 0.        ],
                                     [1.        , 0.        , 0.        ],
                                     [1.        , 0.        , 0.        ],
                                     [0.55422893, 0.83236428, 0.        ],
                                     [0.63035731, 0.        , 0.77630514]]
    # pylint: enable=bad-whitespace

    expected_scores = [np.dot(expected_term_document_matrix[0], row)
                       for row in expected_term_document_matrix]

    self.assertLen(sorted_scores, len(expected_scores))
    for actual_score, expected_score in zip(sorted_scores, expected_scores):
      self.assertAlmostEqual(actual_score, expected_score)

  @parameterized.named_parameters(
      ('0', 0), ('1', 1), ('2', 2), ('5', 5), ('10', 10))
  def test_get_operation_multiplier_respects_max_num_prioritized(
      self, max_num_prioritized):
    # Note: More operations may be prioritized if they are chosen because they
    # are explicitly listed in the description.
    handler = tfidf_handler.TfidfDescriptionHandler(
        max_num_prioritized=max_num_prioritized,
        min_tfidf_score=0)

    benchmark = benchmark_module.Benchmark(
        examples=[benchmark_module.Example(inputs=[1], output=1)],
        description='Tiling a tensor multiple times')
    multipliers = handler.get_operation_multipliers(
        benchmark, settings=settings_module.default_settings())

    actual_num_prioritized = sum(multiplier < 1
                                 for multiplier in multipliers.values())
    self.assertEqual(actual_num_prioritized, max_num_prioritized)
    # All multipliers must be in (0, 1] (no operation is deprioritized).
    self.assertTrue(all(0 < multiplier <= 1
                        for multiplier in multipliers.values()))

  @parameterized.named_parameters(
      ('0', 0.0), ('0_1', 0.1), ('0_2', 0.2), ('1_0', 1.0))
  def test_get_operation_multipliers_respects_min_tfidf_score(
      self, min_tfidf_score):
    # Note: An operation may violate the `min_tfidf_score` threshold if it is
    # chosen because it is explicitly listed in the description.
    handler = tfidf_handler.TfidfDescriptionHandler(
        max_num_prioritized=1000000,
        min_tfidf_score=min_tfidf_score)

    description = 'Tiling a tensor multiple times'
    scores = handler.score_description(description)
    benchmark = benchmark_module.Benchmark(
        examples=[benchmark_module.Example(inputs=[1], output=1)],
        description=description)
    multipliers = handler.get_operation_multipliers(
        benchmark, settings=settings_module.default_settings())

    prioritized_names = [name for name in multipliers.keys()
                         if multipliers[name] < 1]
    expected_prioritized_names = [name for name in scores.keys()
                                  if scores[name] >= min_tfidf_score]
    self.assertCountEqual(prioritized_names, expected_prioritized_names)
    # All multipliers must be in (0, 1] (no operation is deprioritized).
    self.assertTrue(all(0 < multiplier <= 1
                        for multiplier in multipliers.values()))

  @parameterized.named_parameters(
      ('gather', 'gather', [
          'tf.gather(params, indices)',
          'tf.gather(params, indices, axis, batch_dims)',
      ], [
          'tf.gather_nd(params, indices)',
          'tf.gather_nd(params, indices, batch_dims)',
      ]),
      ('gather_nd', 'use tf.gather_nd(params, indices)', [
          'tf.gather_nd(params, indices)',
          'tf.gather_nd(params, indices, batch_dims)',
      ], [
          'tf.gather(params, indices)',
          'tf.gather(params, indices, axis, batch_dims)',
      ]),
      ('alias', 'tf.math.add', [
          'tf.add(x, y)',
      ], [
          'tf.math.add(x, y)',  # This is not an operation name (wrong alias).
      ]),
      ('multiple', 'divide_no_nan and then reduce_max', [
          'tf.math.divide_no_nan(x, y)',
          'tf.reduce_max(input_tensor)',
          'tf.reduce_max(input_tensor, axis)',
          'tf.sparse.reduce_max(sp_input, axis, output_is_sparse)',
      ], [
          'tf.divide(x, y)',
          'tf.reduce_min(input_tensor)',
      ]),
      ('heuristic_is_not_perfect', 'add padding', [
          'tf.add(x, y)',
      ], [
          "tf.pad(tensor, paddings, mode='CONSTANT')",
          "tf.pad(tensor, paddings, mode='CONSTANT', constant_values)",
      ]))
  def test_get_listed_operations_finds_correct_ops(
      self, description, should_find, should_not_find):
    handler = tfidf_handler.TfidfDescriptionHandler()
    listed = handler._get_listed_operations(description)
    for name in should_find:
      self.assertIn(name, listed)
    for name in should_not_find:
      self.assertNotIn(name, listed)

  def test_repr(self):
    handler = tfidf_handler.TfidfDescriptionHandler(max_num_prioritized=12,
                                                    min_tfidf_score=0.34,
                                                    multiplier=0.75)
    self.assertEqual(repr(handler),
                     'TfidfDescriptionHandler(max_num_prioritized=12, '
                     'min_tfidf_score=0.34, multiplier=0.75)')

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for description_handler.py."""

from absl.testing import absltest
from absl.testing import parameterized
from tf_coder.benchmarks import benchmark as benchmark_module
from tf_coder.benchmarks import test_benchmarks
from tf_coder.natural_language import description_handler
from tf_coder.value_search import python_operations
from tf_coder.value_search import value_search_settings as settings_module


class NoChangeDescriptionHandlerTest(absltest.TestCase):

  def setUp(self):
    super(NoChangeDescriptionHandlerTest, self).setUp()
    self.indexing_operation = python_operations.IndexingOperation()
    self.settings = settings_module.default_settings()

  def test_init_succeeds(self):
    description_handler.NoChangeDescriptionHandler()

  def test_init_raises_if_duplicate_name(self):
    with self.assertRaises(ValueError):
      description_handler.NoChangeDescriptionHandler(
          operations=[self.indexing_operation, self.indexing_operation])

  def test_attributes_providing_operations(self):
    handler = description_handler.NoChangeDescriptionHandler(
        operations=[self.indexing_operation])
    self.assertLen(handler.operations, 1)
    self.assertEqual(handler.all_names, [self.indexing_operation.name])

  def test_attributes_default_operations(self):
    handler = description_handler.NoChangeDescriptionHandler()
    self.assertGreater(len(handler.operations), 1)
    self.assertLen(handler.operations, len(handler.all_names))

  def test_get_operation_multipliers(self):
    handler = description_handler.NoChangeDescriptionHandler()
    multipliers = handler.get_operation_multipliers(test_benchmarks.test_add(),
                                                    self.settings)
    self.assertEmpty(multipliers)

  def test_repr(self):
    handler = description_handler.NoChangeDescriptionHandler()
    self.assertEqual(repr(handler), 'NoChangeDescriptionHandler')


class FunctionNameDescriptionHandlerTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('exact_match', 'Use reduce_max', 'tf.reduce_max(input_tensor)', 0.75),
      ('parts_exist', 'Reduce to the max element',
       'tf.reduce_max(input_tensor)', 0.75),
      ('handle_dots', 'Add a sparse tensor', 'tf.sparse.add(a, b)', 0.75),
      ('partial_match', 'Reduce to the min element',
       'tf.reduce_max(input_tensor)', 1.0),
      ('unrelated', 'Reduce to the min element', 'tf.square(x)', 1.0),
      ('not_tf', 'Reduce to the max element', 'IndexingOperation', 1.0))
  def test_get_operation_multipliers(self, description, operation_name,
                                     expected_multiplier):
    handler = description_handler.FunctionNameDescriptionHandler()
    benchmark = benchmark_module.Benchmark(
        examples=[
            benchmark_module.Example(
                inputs=[
                    [10],
                    [20],
                ],
                output=[30],
            ),
        ],
        constants=[],
        description=description,
        target_program='',
        source='test',
        name='test_benchmark')
    multipliers = handler.get_operation_multipliers(
        benchmark, settings_module.default_settings())
    self.assertEqual(multipliers.get(operation_name, 1.0), expected_multiplier)
    # Nothing is deprioritized.
    self.assertTrue(all(0 < value < 1 for value in multipliers.values()))


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for benchmark.py."""

from absl.testing import absltest
from absl.testing import parameterized
from tf_coder.benchmarks import benchmark
from tf_coder.benchmarks import test_benchmarks


class BenchmarkTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('no_examples', []),
      ('none_output', [benchmark.Example(inputs=[[1]], output=None)]),
      ('inconsistent_num_inputs', [
          benchmark.Example(inputs=[[1]], output=[1]),  # 1 input.
          benchmark.Example(inputs=[[1], [2]], output=[3]),  # 2 inputs.
      ]))
  def test_benchmark_init_raises(self, examples):
    with self.assertRaises(ValueError):
      benchmark.Benchmark(examples, name='test')

  def test_benchmark_init_computes_num_inputs(self):
    examples = [
        benchmark.Example(
            inputs=[
                [1],
                [2],
            ],
            output=[1, 2]
        ),
        benchmark.Example(
            inputs=[
                [[1, 2], [3, 4]],
                [[5, 6], [7, 8]],
            ],
            output=[[1, 2], [3, 4], [5, 6], [7, 8]]
        ),
    ]
    self.assertEqual(benchmark.Benchmark(examples, name='test').num_inputs, 2)

  def test_benchmark_init_assigns_name_if_not_given(self):
    benchmark_1 = benchmark.Benchmark([benchmark.Example(inputs=[1], output=1)])
    benchmark_2 = benchmark.Benchmark([benchmark.Example(inputs=[2], output=2)])
    self.assertEqual(benchmark_1.name, 'unnamed_benchmark_1')
    self.assertEqual(benchmark_2.name, 'unnamed_benchmark_2')

  def test_benchmark_not_ignored_by_default(self):
    not_ignored_list = [test_benchmarks.test_add(), test_benchmarks.test_cast()]
    for not_ignored in not_ignored_list:
      self.assertFalse(not_ignored.should_ignore)
      self.assertIsNone(not_ignored.ignore_reason)

  def test_ignore_decorator(self):
    ignored_benchmark = test_benchmarks.inconsistent_target_program()
    self.assertTrue(ignored_benchmark.should_ignore)
    self.assertIsNotNone(ignored_benchmark.ignore_reason)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Runs value search to get a dataset of I/O pairs with expressions."""

import collections
import copy
import enum
import itertools
import random
import typing
from typing import Any, Dict, List, NamedTuple, Optional, Text, Tuple, Union

from absl import logging
import six
import tensorflow as tf
from tf_coder import tf_coder_utils
from tf_coder.value_search import all_operations
from tf_coder.value_search import operation_base
from tf_coder.value_search import value as value_module


IOExample = NamedTuple('IOExample',
                       [('expression', Text),
                        ('input_values', List[value_module.Value]),
                        ('output_value', value_module.Value),
                        ('num_inputs', int),
                        ('operations', List[operation_base.Operation])])

NEW_INPUT_NAME = 'NEW_INPUT'

EPSILON = 1e-8

# Bucket boundaries for nonnegative integers. Anything greater than or equal to
# 1000 will be placed in the last bucket, so the number of buckets equals the
# length of the list.
COUNT_BOUNDARIES = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
                    12, 15, 20, 25, 50, 75, 100, 200, 500, 1000]

# Bucket boundaries for arbitrary floats.
_POSITIVE_BOUNDARIES = [
    EPSILON, 0.01, 0.1, 0.5, 1, 2, 5, 10, 20, 50, 100, 200, 500, 1000]
FLOAT_BOUNDARIES = ([-float('inf')] + [-x for x in _POSITIVE_BOUNDARIES[::-1]] +
                    _POSITIVE_BOUNDARIES)

# The max rank for tensor featurization.
MAX_RANK = 4


def extract_values_with_collapsed_subtrees(
    value: value_module.Value) -> List[value_module.Value]:
  """Collapses subtrees, see docstring of extract_examples_from_value."""
  if not isinstance(value, value_module.OperationValue):
    # Base case: the value is a leaf.
    return [value]

  value = typing.cast(value_module.OperationValue, value)

  results = []  # type: List[value_module.Value]

  # Each OperationApplication namedtuple represents one way of reaching this
  # Value by applying an operation to some arguments. Choose one at random.
  # If we recursed through all possible choices, we would heavily bias the
  # dataset toward Values with many possible expressions, e.g., those using many
  # commutative ops.
  operation_application = random.choice(value.operation_applications)

  children_possibilities = [extract_values_with_collapsed_subtrees(child)
                            for child in operation_application.arg_values]
  for children_choices in itertools.product(*children_possibilities):
    results.append(value_module.OperationValue(
        value=value.value,
        operation=operation_application.operation,
        arg_values=children_choices))

  # Include the case where the entire AST is replaced with a new input. Skip
  # tensor conversion, meaning don't convert a tuple of tensors into a single
  # tensor with one greater rank; keep the object as-is.
  results.append(value_module.InputValue(value=value.value,
                                         name=NEW_INPUT_NAME,
                                         skip_tensor_conversion=True))
  return results


def count_num_inputs(
    value: value_module.Value,
    name_counter: Optional[collections.Counter] = None
) -> Tuple[int, collections.Counter]:
  """Counts the number of distinct inputs, where all new inputs are distinct.

  The new inputs are created by extract_values_with_collapsed_subtrees, where
  each collapsed subtree becomes a new input.

  Args:
    value: The Value object to process.
    name_counter: Used for recursion by this function. Clients should ignore
      this argument.

  Returns:
    The number of inputs, and a Counter over the input names.
  """
  if name_counter is None:
    name_counter = collections.Counter()

  if isinstance(value, value_module.InputValue):
    name_counter[value.name] += 1
  elif isinstance(value, value_module.OperationValue):
    if len(value.operation_applications) != 1:
      raise ValueError('Expected exactly one way of constructing this Value.')
    for child in value.operation_applications[0].arg_values:
      count_num_inputs(child, name_counter)

  num_existing_names = len(name_counter)
  if NEW_INPUT_NAME in name_counter:
    num_existing_names -= 1
  num_inputs = num_existing_names + name_counter[NEW_INPUT_NAME]
  return num_inputs, name_counter


def normalize_names_and_extract_values(
    value: value_module.Value,
    name_counter: collections.Counter) -> Dict[Text, value_module.InputValue]:
  """Normalizes names of inputs to be in1, in2, etc., in place.

  Important: This implementation assumes that all InputValue objects in the tree
  have distinct identities, so changing the name of one InputValue will not
  affect other InputValues elsewhere in the tree. This requirement can be met by
  calling `value.copy()` before passing `value` to this function.

  Args:
    value: The Value to process.
    name_counter: A dict containing all existing names as keys, e.g., as
      returned by count_num_inputs.

  Returns:
    A dict mapping input names to InputValue objects.
  """
  renaming_map = {}
  for old_name in sorted(name_counter):
    if old_name == NEW_INPUT_NAME:
      continue  # Handle these later, not via the renaming_map.
    renaming_map[old_name] = 'in{}'.format(len(renaming_map) + 1)

  name_to_value_map = {}

  def recurse(value: value_module.Value, new_name_index: int) -> int:
    """Recursively normalizes names using renaming_map and new_name_index.

    Args:
      value: The Value to recurse on.
      new_name_index: An index for naming the next "new input" encountered,
        which was created via the subtree-collapsing process.

    Returns:
      The updated new_name_index to use in future recursion.
    """
    if isinstance(value, value_module.InputValue):

      if value.name in renaming_map:
        value.name = renaming_map[value.name]
      else:
        assert value.name == NEW_INPUT_NAME
        value.name = 'in{}'.format(new_name_index)
        new_name_index += 1

      if value.name in name_to_value_map:
        assert name_to_value_map[value.name] == value
      else:
        name_to_value_map[value.name] = value

    elif isinstance(value, value_module.OperationValue):
      if len(value.operation_applications) != 1:
        raise ValueError('Expected exactly one way of constructing this Value.')
      for child in value.operation_applications[0].arg_values:
        new_name_index = recurse(child, new_name_index)

    return new_name_index

  recurse(value, len(renaming_map) + 1)

  return name_to_value_map


def extract_operations(
    value: value_module.Value,
    operation_list: Optional[List[operation_base.Operation]] = None
) -> List[operation_base.Operation]:
  """Returns a list of Operations used in the Value."""
  if operation_list is None:
    operation_list = []

  if isinstance(value, value_module.OperationValue):
    operation_list.append(value.operation_applications[0].operation)
    for child in value.operation_applications[0].arg_values:
      extract_operations(child, operation_list)

  return operation_list


def extract_examples_from_value(value: value_module.Value,
                                max_num_inputs: int = 3) -> List[IOExample]:
  """Extracts IOExample namedtuples from a Value.

  A Value represents an expression as well as its value when evaluated. Multiple
  I/O examples can be extracted from a single Value, where different
  subexpressions are considered inputs.

  For an example, we will denote tensors and other "complex" objects with
  floats, and "simple" constants with ints. Suppose that we run value search
  starting with inputs `in1 = 4.0` and `in2 = 3.0`, using a variety of int
  constants including `1` and `2`. During value search we may find a Value
  representing the expression `((1 + in1) * (in2 - 1)) + 2` that evaluates to
  12.0. From this Value, we may extract many examples including:

    * The original expression:
      ((1 + in1) * (in2 - 1)) + 2 -> 12.0, where in1 = 4.0, in2 = 3.0

    * Collapse different subtrees and treat them as new inputs:
      (in1 * (in2 - 1)) + 2 -> 12.0, where in1 = 5.0, in2 = 3.0
      ((1 + in1) * in2) + 2 -> 12.0, where in1 = 4.0, in2 = 2.0
      (in1 * in2) + 2 -> 12.0, where in1 = 5.0, in2 = 2.0
      in1 + 2 -> 12.0, where in1 = 10.0

  We do not turn constants into extra inputs because inputs are most commonly
  Tensors. We do not permute inputs (e.g., turning `tf.add(in1, in2)` into
  `tf.add(in2, in1)`) here because that can easily be done later. This also
  means we do not have to permute inputs within the expression string, which is
  tricky and unnecessary for synthetic data generation.

  Note that all expressions evaluate to the same result as the original Value
  does. This is sufficient because we can extract examples from all Values found
  during value search, which includes all subtrees of the current Value.

  Args:
    value: The starting Value to extract examples from.
    max_num_inputs: The maximum number of inputs for an extracted example. All
      extracted examples will have at least one input.

  Returns:
    A list of IOExample namedtuples.
  """
  values = extract_values_with_collapsed_subtrees(value)
  if values[-1].reconstruct_expression(use_cache=False) == NEW_INPUT_NAME:
    # The last element is the case where the entire expression is turned into a
    # single new input, which is not useful.
    del values[-1]

  # Create IOExamples by filtering out Values with too many inputs and
  # normalizing names.
  examples = []
  for v in values:
    # Copy (recursively) to completely isolate different expressions and
    # different subtrees within one expression.
    v = v.copy()

    num_inputs, name_counter = count_num_inputs(v)
    if not 0 < num_inputs <= max_num_inputs:
      # If `value` contains only constants and no inputs, then the unchanged
      # expression will have 0 inputs, but inputs will be introduced by
      # collapsing subtrees.
      continue
    name_to_value_map = normalize_names_and_extract_values(v, name_counter)
    assert len(name_to_value_map) == num_inputs

    examples.append(IOExample(
        expression=v.reconstruct_expression(use_cache=False),
        input_values=[name_to_value_map[name]
                      for name in sorted(name_to_value_map)],
        output_value=v,
        num_inputs=num_inputs,
        operations=extract_operations(v)))

  return examples


def _bucket(number: float, bucket_boundaries: List[float]) -> int:
  """Returns an int representing a bucket for the number."""
  for index, (left, right) in enumerate(zip(bucket_boundaries,
                                            bucket_boundaries[1:])):
    if left <= number < right:
      return index
  return len(bucket_boundaries) - 1


def _as_tensor_to_featurize(
    value: value_module.Value) -> Union[tf.Tensor, tf.SparseTensor]:
  """Extracts a Tensor or SparseTensor to featurize."""
  to_featurize = None
  if value.is_tensor or value.is_sparse_tensor:
    to_featurize = value.value
  elif value.is_primitive and isinstance(value.value, (int, float, bool)):
    to_featurize = tf.constant(value.value)
  elif value.is_sequence:
    if value.elem_type_is_tensor or value.elem_type_is_sparse_tensor:
      to_featurize = value.value[0]
    else:
      # In order to be a valid Value, a possibly-multidimensional sequence of
      # primitives must be convertible to a tensor.
      to_featurize = tf.constant(value.value)

  if to_featurize is None:
    raise ValueError('Could not featurize value {}'.format(value))

  assert isinstance(to_featurize, (tf.Tensor, tf.SparseTensor))
  return to_featurize


def _elements(tensor: Union[tf.Tensor, tf.SparseTensor]) -> tf.Tensor:
  """Returns a Tensor of non-default elements in a Tensor or SparseTensor."""
  return tensor.values if isinstance(tensor, tf.SparseTensor) else tensor


def _elements_list(tensor: Union[tf.Tensor, tf.SparseTensor]) -> List[float]:
  """Returns a list of non-default elements in a Tensor or SparseTensor."""
  return list(map(float, tf.reshape(_elements(tensor), [-1]).numpy()))


def _safe_divide(numerator: int, denominator: int) -> float:
  """Normal division except anything/0 becomes 0.0."""
  return numerator / denominator if denominator else 0.0


def _fixed_length_shape(tensor) -> List[int]:
  """Returns the tensor shape, truncated or padded with 0 to length MAX_RANK."""
  fixed_length_shape = list(tensor.shape[:MAX_RANK])
  fixed_length_shape += [0] * (MAX_RANK - len(fixed_length_shape))
  return fixed_length_shape


def featurize_value(
    value: value_module.Value) -> Dict[Text, Union[List[float], List[Text]]]:
  """Returns a dict that featurizes a Value.

  The feature dict will contain the following:

   * 'kind': One int denoting whether the value is a primitive, sequence,
       Tensor, SparseTensor, or other.
   * 'dtype': One int denoting the DType of the value.
   * 'rank': One int in the range [0, MAX_RANK], where MAX_RANK denotes any rank
       greater than or equal to MAX_RANK.
   * 'shape': A list of MAX_RANK ints denoting the shape, 0-padded.
   * 'shape_buckets': Like 'shape' but bucketed.
   * 'floats': Stats about the elements as raw floats.
   * 'float_buckets': Like 'floats' but bucketed.
   * 'counts': Numbers of elements with various properties.
   * 'count_buckets': Like 'counts' but bucketed.
   * 'fractions': Fractions of elements with various properties.
   * 'booleans': Boolean properties of the value, as 0-1 ints.
   * 'value_string': The input as a string.

  Args:
    value: The Value to featurize.
  """
  to_featurize = _as_tensor_to_featurize(value)
  features = {}  # type: Dict[Text, Union[List[float], List[Text]]]

  features['kind'] = [(
      0 if value.is_primitive else  # pylint: disable=g-long-ternary
      1 if value.is_sequence else  # pylint: disable=g-long-ternary
      2 if value.is_tensor else
      3 if value.is_sparse_tensor else
      4)]

  supported_dtypes = (
      tf_coder_utils.INT_DTYPES + tf_coder_utils.FLOAT_DTYPES + (tf.bool,))
  if to_featurize.dtype not in supported_dtypes:
    raise ValueError('Cannot featurize unsupported dtype: {}'.format(
        to_featurize.dtype))
  features['dtype'] = [supported_dtypes.index(to_featurize.dtype)]

  features['rank'] = [min(len(to_featurize.shape), MAX_RANK)]

  features['shape'] = _fixed_length_shape(to_featurize)
  features['shape_buckets'] = [
      _bucket(dimension_length, COUNT_BOUNDARIES)
      for dimension_length in features['shape']]

  # "Elements" are the provided values (without default elements in
  # SparseTensors). Cast elements to float32 because, e.g., we can't compute the
  # mean of a bool tensor.
  elements = tf.cast(_elements(to_featurize), tf.float32)
  elements_list = _elements_list(to_featurize)

  max_value = tf_coder_utils.max_tensor_value(elements)
  min_value = tf_coder_utils.min_tensor_value(elements)
  mean_value = float(tf.reduce_mean(elements))
  mean_magnitude = float(tf.reduce_mean(tf.abs(elements)))
  features['floats'] = [max_value, min_value, mean_value, mean_magnitude]
  features['float_buckets'] = [_bucket(f, FLOAT_BOUNDARIES)
                               for f in features['floats']]

  # The total size of the tensor, meaning the product of dimension lengths. This
  # may be huge for SparseTensors.
  size = tf_coder_utils.num_tensor_elements(to_featurize)
  # For SparseTensors, num_elements != size.
  num_elements = len(elements_list)

  positive = int(tf.reduce_sum(tf.cast(elements > 0, tf.int32)))
  zeros = int(tf.reduce_sum(tf.cast(
      tf.abs(tf.cast(elements, tf.float32)) < EPSILON, tf.int32)))
  ones = int(tf.reduce_sum(tf.cast(
      tf.abs(tf.cast(elements, tf.float32) - 1.0) < EPSILON, tf.int32)))
  negative = int(tf.reduce_sum(tf.cast(elements < 0, tf.int32)))
  probabilities = int(tf.reduce_sum(tf.cast(
      tf.logical_and(0 <= elements, elements <= 1), tf.int32)))
  unique = len(set(elements_list))

  counts = [size, num_elements, positive, zeros, ones, negative, probabilities,
            unique]

  features['counts'] = counts
  features['count_buckets'] = [_bucket(c, COUNT_BOUNDARIES)
                               for c in features['counts']]

  assert counts[0] == size and counts[1] == num_elements
  features['fractions'] = [_safe_divide(num_elements, size)]
  features['fractions'].extend([_safe_divide(c, num_elements)
                                for c in counts[2:]])

  # TODO(kshi): Consider computing the number of times the mode appears.

  # TODO(kshi): Consider computing, for each axis, whether that axis represents
  # a probability distribution, with all elements in [0, 1] summing to 1.

  is_sorted = (int(tf.reduce_all(tf.equal(elements, tf.sort(elements))))
               if len(elements.shape) else 1)
  is_finite = int(tf.reduce_all(tf.math.is_finite(elements)))
  all_positive = int(positive == num_elements)
  all_nonnegative = int(positive + zeros == num_elements)
  all_negative = int(negative == num_elements)
  all_zero_one = int(zeros + ones == num_elements)
  all_probabilities = int(probabilities == num_elements)
  all_unique = int(unique == num_elements)

  features['booleans'] = [
      is_sorted, is_finite, all_positive, all_nonnegative, all_negative,
      all_zero_one, all_probabilities, all_unique]

  features['value_string'] = [repr(value)]

  return features


def _comparison(a, b) -> int:
  """Returns an int in [0, 2] representing the comparison result."""
  return 0 if a < b else 1 if a == b else 2


def featurize_input_and_output(
    input_value: value_module.Value,
    output_value: value_module.Value) -> Dict[Text, List[float]]:
  """Returns a dict that featurizes relationships between an input and output.

  The feature dict will have the following keys (all start with 'io_'):

   * 'io_comparisons', list[int]: Ints in the range [0, 2] denoting which is
       larger according to various metrics.
   * 'io_counts', list[int]: Numbers of elements with various properties.
   * 'io_count_buckets', list[int]: Like 'io_counts' but bucketed.
   * 'io_fractions', list[float]: Fractions of elements with various properties.
   * 'io_booleans', list[int]: Ints in the range [0, 1] denoting various boolean
       properties.

  Args:
    input_value: The input Value.
    output_value: The output Value.
  """
  input_tensor = _as_tensor_to_featurize(input_value)
  output_tensor = _as_tensor_to_featurize(output_value)

  features = {}  # type: Dict[Text, List[float]]

  input_size = tf_coder_utils.num_tensor_elements(input_tensor)
  output_size = tf_coder_utils.num_tensor_elements(output_tensor)

  input_elements = _elements_list(input_tensor)
  output_elements = _elements_list(output_tensor)
  input_num_elements = len(input_elements)
  output_num_elements = len(output_elements)

  input_rank = len(input_tensor.shape)
  output_rank = len(output_tensor.shape)

  input_shape = _fixed_length_shape(input_tensor)
  output_shape = _fixed_length_shape(output_tensor)

  features['io_comparisons'] = [
      _comparison(input_size, output_size),
      _comparison(input_num_elements, output_num_elements),
      _comparison(input_rank, output_rank),
  ]
  features['io_comparisons'].extend(
      [_comparison(input_length, output_length)
       for input_length, output_length in zip(input_shape, output_shape)])

  # Count elements appearing in both the input and output.
  input_elements_set = set(input_elements)
  output_elements_set = set(output_elements)

  inputs_in_output = sum(e in output_elements_set for e in input_elements)
  outputs_in_input = sum(e in input_elements_set for e in output_elements)
  num_unique_overlaps = len(input_elements_set & output_elements_set)

  features['io_counts'] = [inputs_in_output, outputs_in_input,
                           num_unique_overlaps]
  features['io_count_buckets'] = [_bucket(c, COUNT_BOUNDARIES)
                                  for c in features['io_counts']]
  features['io_fractions'] = [
      _safe_divide(inputs_in_output, input_num_elements),
      _safe_divide(outputs_in_input, output_num_elements),
      _safe_divide(num_unique_overlaps, input_num_elements),
      _safe_divide(num_unique_overlaps, output_num_elements),
  ]

  all_inputs_in_output = int(inputs_in_output == len(input_elements))
  all_outputs_in_input = int(outputs_in_input == len(output_elements))
  features['io_booleans'] = [
      int(input_tensor.shape.as_list() == output_tensor.shape.as_list()),
      int(input_tensor.dtype == output_tensor.dtype),
      all_inputs_in_output,
      all_outputs_in_input,
  ]
  features['io_booleans'].extend([
      int(input_length in output_shape) if input_length else 0
      for input_length in input_shape])
  features['io_booleans'].extend([
      int(output_length in input_shape) if output_length else 0
      for output_length in output_shape])

  return features


def create_examples(io_example: IOExample,
                    max_num_inputs: int = 3,
                    permute_inputs: bool = True) -> List[Dict[Text, Any]]:
  """Creates example dicts for the I/O example."""
  examples = []
  operation_list = all_operations.get_operations(include_sparse_operations=True)
  operation_counter = collections.Counter(
      [op.name for op in io_example.operations])
  operation_counts = [operation_counter[op.name] for op in operation_list]

  num_inputs = min(len(io_example.input_values), max_num_inputs)

  try:
    output_features = featurize_value(io_example.output_value)
    input_features = []
    for input_value in io_example.input_values[:max_num_inputs]:
      combined_features = featurize_value(input_value)
      combined_features.update(
          featurize_input_and_output(input_value, io_example.output_value))
      input_features.append(combined_features)

    dummy_value = value_module.ConstantValue(0)
    dummy_input_features = featurize_value(dummy_value)
    dummy_input_features.update(
        featurize_input_and_output(dummy_value, dummy_value))

  except ValueError as e:
    logging.warning('%s: could not featurize IOExample %s', e, io_example)
    return []

  permutations = (itertools.permutations(range(num_inputs)) if permute_inputs
                  else [list(range(num_inputs))])
  for permutation in permutations:

    feature_dict = collections.defaultdict(list)
    feature_dict['num_inputs'] = [num_inputs]
    feature_dict.update(copy.deepcopy(output_features))

    padded_input_features = [input_features[index] for index in permutation]
    for _ in range(max_num_inputs - num_inputs):
      padded_input_features.append(dummy_input_features)

    for input_features_dict in padded_input_features:
      for key, value in six.iteritems(input_features_dict):
        feature_dict[key].extend(value)

    feature_dict['operations'] = operation_counts
    feature_dict['expression'] = [io_example.expression]

    examples.append(feature_dict)

  return examples


def _bytes_feature(strings):
  """Returns a bytes_list from a list of strings."""
  return tf.train.Feature(bytes_list=tf.train.BytesList(
      value=[s.encode('utf-8') for s in strings]))


def _int64_feature(ints):
  """Returns an int64_list from a list of ints."""
  return tf.train.Feature(int64_list=tf.train.Int64List(value=ints))


def _float_feature(floats):
  """Returns a float_list from a list of numbers."""
  return tf.train.Feature(float_list=tf.train.FloatList(value=floats))


@enum.unique
class FeatureType(enum.Enum):
  """Types of features in the tf.train.Example proto."""
  # TODO(kshi): Can the values be _int64_feature etc.? Pytype doesn't seem to
  # understand that correctly -- this may be a pytype bug.
  INT = 1
  FLOAT = 2
  STRING = 3


FEATURE_NAME_TO_TYPE = {
    # Features from featurize_value.
    'kind': FeatureType.INT,
    'dtype': FeatureType.INT,
    'rank': FeatureType.INT,
    'shape': FeatureType.INT,
    'shape_buckets': FeatureType.INT,
    'floats': FeatureType.FLOAT,
    'float_buckets': FeatureType.INT,
    'counts': FeatureType.INT,
    'count_buckets': FeatureType.INT,
    'fractions': FeatureType.FLOAT,
    'booleans': FeatureType.INT,
    'value_string': FeatureType.STRING,

    # Features from featurize_input_and_output.
    'io_comparisons': FeatureType.INT,
    'io_counts': FeatureType.INT,
    'io_count_buckets': FeatureType.INT,
    'io_fractions': FeatureType.FLOAT,
    'io_booleans': FeatureType.INT,

    # Features added in create_examples.
    'num_inputs': FeatureType.INT,
    'operations': FeatureType.INT,
    'expression': FeatureType.STRING,
}


def create_tf_examples(io_example: IOExample,
                       max_num_inputs: int = 3,
                       permute_inputs: bool = True) -> List[tf.train.Example]:
  """Creates tf.train.Example protos for the I/O example."""
  tf_examples = []
  for example in create_examples(io_example,
                                 max_num_inputs=max_num_inputs,
                                 permute_inputs=permute_inputs):
    feature_dict = {}
    assert set(FEATURE_NAME_TO_TYPE.keys()) == set(example.keys())

    for feature_name, feature_value in six.iteritems(example):
      feature_type = FEATURE_NAME_TO_TYPE[feature_name]
      conversion_function = (
          _int64_feature if feature_type == FeatureType.INT else  # pylint: disable=g-long-ternary
          _float_feature if feature_type == FeatureType.FLOAT else
          _bytes_feature if feature_type == FeatureType.STRING else
          None)
      assert conversion_function is not None
      feature_dict[feature_name] = conversion_function(feature_value)

    tf_examples.append(tf.train.Example(features=tf.train.Features(
        feature=feature_dict)))

  return tf_examples
